from python_helper import SettingHelper, StringHelper, Constant, log, EnvironmentHelper, ObjectHelper, Test

LOG_HELPER_SETTINGS = {
    log.LOG : False,
    log.SUCCESS : True,
    log.SETTING : True,
    log.DEBUG : True,
    log.WARNING : True,
    log.WRAPPER : True,
    log.FAILURE : True,
    log.ERROR : True,
    log.TEST : False
}

@Test(
    environmentVariables={
        SettingHelper.ACTIVE_ENVIRONMENT : None,
        **LOG_HELPER_SETTINGS
    }
)
def updateActiveEnvironment_withSuccess() :
    # Arrange
    originalACTIVE_ENVIRONMENT_VALUE = SettingHelper.ACTIVE_ENVIRONMENT_VALUE
    originalGottenActiveEnvironment = SettingHelper.getActiveEnvironment()
    originalActiveEnvironmentIsDefault = SettingHelper.activeEnvironmentIsDefault()
    originalActiveEnvironment = EnvironmentHelper.get(SettingHelper.ACTIVE_ENVIRONMENT)
    originalACTIVE_ENVIRONMENT_VALUEAfterSettingAnotherOne = SettingHelper.ACTIVE_ENVIRONMENT_VALUE
    myNewActiveEnvironment = 'my new artive environment'

    # Act
    myGottenNewActiveEnvironment = SettingHelper.updateActiveEnvironment(myNewActiveEnvironment)

    # Assert
    assert SettingHelper.DEFAULT_ENVIRONMENT == originalGottenActiveEnvironment
    assert SettingHelper.DEFAULT_ENVIRONMENT == originalActiveEnvironment
    assert True == originalActiveEnvironmentIsDefault
    assert myNewActiveEnvironment == EnvironmentHelper.get(SettingHelper.ACTIVE_ENVIRONMENT)
    assert False == SettingHelper.activeEnvironmentIsDefault()
    assert myNewActiveEnvironment == SettingHelper.getActiveEnvironment()
    assert ObjectHelper.isNotEmpty(myGottenNewActiveEnvironment)
    assert myGottenNewActiveEnvironment == myNewActiveEnvironment
    assert SettingHelper.DEFAULT_ENVIRONMENT == originalACTIVE_ENVIRONMENT_VALUE
    assert SettingHelper.DEFAULT_ENVIRONMENT == originalACTIVE_ENVIRONMENT_VALUEAfterSettingAnotherOne
    assert SettingHelper.ACTIVE_ENVIRONMENT_VALUE == myNewActiveEnvironment

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS : True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        'MY_COMPLEX_ENV' : ' -- my complex value -- ',
        'LATE_VALUE' : '-- late environment value --',
        'ONLY_ENVIRONMENT_VARIABLE' : 'only environment variable value',
        **LOG_HELPER_SETTINGS
    }
)
def mustReadSettingFile() :
    # Arrange
    settingFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource','application.yml'])

    # Act
    readdedSettingTree = SettingHelper.getSettingTree(settingFilePath, keepDepthInLongString=True)
    # log.prettyPython(mustReadSettingFile, 'readdedSettingTree', readdedSettingTree)

    # Assert
    assert 'self reference value' == SettingHelper.getSetting('my.self-reference-key', readdedSettingTree)
    assert 'other self reference value as well' == SettingHelper.getSetting('my.other.self-reference-key.as-well', readdedSettingTree)
    assert 'other repeated self reference value as well' == SettingHelper.getSetting('my.other.repeated.self-reference-key.as-well', readdedSettingTree)
    assert 'my default value' == SettingHelper.getSetting('my.configuration-without-environment-variable-key', readdedSettingTree)
    assert "my default value" == SettingHelper.getSetting('my.configuration-without-environment-variable-key-with-value-surrounded-by-single-quotes', readdedSettingTree)
    assert 'my default value' == SettingHelper.getSetting('my.configuration-without-environment-variable-key-and-space-after-colon', readdedSettingTree)
    assert 'self reference value' == SettingHelper.getSetting('my.configuration', readdedSettingTree)
    assert 'self reference value' == SettingHelper.getSetting('my.own.configuration', readdedSettingTree)
    assert 'other root value' == SettingHelper.getSetting('other.root.key', readdedSettingTree)
    assert 'other root value' == SettingHelper.getSetting('my.own.very.deep.configuration', readdedSettingTree)
    assert 'other self reference value as well' == SettingHelper.getSetting('my.other-with-other-name.self-reference-key.as-well', readdedSettingTree)
    assert 'self reference value' == SettingHelper.getSetting('my.other-with-other-name.configuration', readdedSettingTree)
    assert 'other self reference value as well' == SettingHelper.getSetting('my.other-with-other-name.configuration-as-well', readdedSettingTree)
    assert 'other repeated self reference value as well' == SettingHelper.getSetting('my.other-with-other-name.configuration-repeated-as-well', readdedSettingTree)
    assert SettingHelper.getSetting('my.override-case.overridden', readdedSettingTree) is None
    assert 'overrider configuration' == SettingHelper.getSetting('my.override-case.overrider', readdedSettingTree)

    assert 'delayed assignment value' == SettingHelper.getSetting('some-reference.before-its-assignment', readdedSettingTree)
    assert 'delayed assignment value' == SettingHelper.getSetting('some-reference.much.before-its-assignment', readdedSettingTree)
    assert "'''  value  ''' with spaces" == SettingHelper.getSetting('some-key.with-an-enter-in-between-the-previous-one', readdedSettingTree)
    assert f'''Hi
                every
            one''' == SettingHelper.getSetting('long.string', readdedSettingTree)
    assert f'''Hi
                            every
                            one
                            this
                            is
                            the
                            deepest
                            long
                                        string
                            here''' == SettingHelper.getSetting('deepest.long.string.ever.long.string', readdedSettingTree)
    assert f'''me
                    being
        not
                    fshds''' == SettingHelper.getSetting('not.idented.long.string', readdedSettingTree)
    assert 'abcdefg' == SettingHelper.getSetting('it.contains.one-setting-injection', readdedSettingTree)
    assert 'abcdefghijklm' == SettingHelper.getSetting('it.contains.two-consecutive-setting-injection', readdedSettingTree)
    assert 'abcdefghijklm' == SettingHelper.getSetting('it.contains.one-inside-of-the-other-setting-injection', readdedSettingTree)
    assert 'ABCD-- my complex value --EFG' == SettingHelper.getSetting('it.contains.one-setting-injection-with-environment-variable', readdedSettingTree)
    assert 'ABCDEFGEFG-- my complex value --HIJKLMNOP' == SettingHelper.getSetting('it.contains.one-inside-of-the-other-setting-injection-with-environment-variable', readdedSettingTree)
    assert 'abcdefghijklm' == SettingHelper.getSetting('it.contains.two-consecutive-setting-injection-with-missing-environment-variable', readdedSettingTree)
    assert 'abcd-- late value ----abcd---- late value ----abcd--efg' == SettingHelper.getSetting('it.contains.some-composed-key.pointing-to.a-late-value', readdedSettingTree)
    assert 'abcd-- late environment value ----abcd--it.contains.late-value--abcd--efg' == SettingHelper.getSetting('it.contains.some-composed-key.pointing-to.a-late-value-with-an-environment-variable-in-between', readdedSettingTree)
    assert '-- late value --' == SettingHelper.getSetting('it.contains.late-value', readdedSettingTree)
    assert 'only environment variable value' == SettingHelper.getSetting('it.contains.environment-variable.only', readdedSettingTree)
    assert 'ABCD -- only environment variable value -- EFGH' == SettingHelper.getSetting('it.contains.environment-variable.surrounded-by-default-values', readdedSettingTree)
    assert 'ABCD -- "some value followed by: "only environment variable value\' and some following default value\' -- EFGH' == SettingHelper.getSetting('it.contains.environment-variable.in-between-default-values', readdedSettingTree)
    assert 'ABCD -- very late definiton value -- EFGH' == SettingHelper.getSetting('it.contains.refference.to-a-late-definition', readdedSettingTree)
    assert 222233444 == SettingHelper.getSetting('handle.integer', readdedSettingTree)
    assert 2.3 == SettingHelper.getSetting('handle.float', readdedSettingTree)
    assert True == SettingHelper.getSetting('handle.boolean', readdedSettingTree)
    assert 222233444 == SettingHelper.getSetting('handle.late.integer', readdedSettingTree)
    assert 2.3 == SettingHelper.getSetting('handle.late.float', readdedSettingTree)
    assert True == SettingHelper.getSetting('handle.late.boolean', readdedSettingTree)
    assert [] == SettingHelper.getSetting('handle.empty.list', readdedSettingTree)
    assert {} == SettingHelper.getSetting('handle.empty.dictionary-or-set', readdedSettingTree)
    assert (()) == SettingHelper.getSetting('handle.empty.tuple', readdedSettingTree)
    assert 'ABCD -- 222233444 -- EFGH' == SettingHelper.getSetting('some-not-string-selfreference.integer', readdedSettingTree)
    assert 'ABCD -- 2.3 -- EFGH' == SettingHelper.getSetting('some-not-string-selfreference.float', readdedSettingTree)
    assert 'ABCD -- True -- EFGH' == SettingHelper.getSetting('some-not-string-selfreference.boolean', readdedSettingTree)

@Test(
    environmentVariables={
        **{},
        **LOG_HELPER_SETTINGS
    }
)
def mustNotReadSettingFile() :
    # Arrange
    settingFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource','application-circular-reference.yml'])
    circularReferenceSettingInjectionDictionary = {
        'circular.reference.on.key': {
            'SETTING_KEY': 'key',
            'SETTING_VALUE': '${circular.reference.on.other-key}',
            'SETTING_NODE_KEY': 'circular.reference.on'
        },
        'circular.reference.on.other-key': {
            'SETTING_KEY': 'other-key',
            'SETTING_VALUE': '${circular.reference.on.key}',
            'SETTING_NODE_KEY': 'circular.reference.on'
        },
        'circular.key': {
            'SETTING_KEY': 'key',
            'SETTING_VALUE': '${circular.other-key}',
            'SETTING_NODE_KEY': 'circular'
        },
        'circular.other-key': {
            'SETTING_KEY': 'other-key',
            'SETTING_VALUE': '${circular.key}',
            'SETTING_NODE_KEY': 'circular'
        }
    }
    fallbackSettingTree = None
    settingTree = {
        'circular': {
            'reference': {
                'on': {}
            }
        }
    }
    exceptionMessage = f'Circular reference detected in following setting injections: {StringHelper.prettyPython(circularReferenceSettingInjectionDictionary)}{Constant.NEW_LINE}FallbackSettingTree: {StringHelper.prettyPython(fallbackSettingTree)}{Constant.NEW_LINE}SettingTree: {StringHelper.prettyPython(settingTree)}'

    # Act
    readdedSettingTree = {}
    exception = None
    try :
        readdedSettingTree = SettingHelper.getSettingTree(settingFilePath, keepDepthInLongString=True)
    except Exception as ext :
        exception = ext

    # Assert
    assert {} == readdedSettingTree
    assert exceptionMessage == str(exception), f'assert {exceptionMessage} == {str(exception)}'

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS : True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **LOG_HELPER_SETTINGS
    }
)
def mustPrintSettingTree() :
    # Arrange
    settingFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource','application.yml'])
    readdedSettingTree = SettingHelper.getSettingTree(settingFilePath, keepDepthInLongString=True)

    # Act
    SettingHelper.printSettings(readdedSettingTree,'some name')

    # Assert
    assert ObjectHelper.isNotNone(readdedSettingTree)

def querySetting_withSuccess() :
    # Arrange
    tree = {
        'key' : {
            'key' : {
                'some-query-key' : {
                    'key' : {
                        'key' : {
                            'some-query-key' : 'value',
                            'key' : 'value'
                        }
                    },
                    'other-key' : {
                        'key' : {
                            'other-key' : 'value',
                            'key' : 'value'
                        }
                    }
                },
                'key' : {
                    'key' : {
                        'key' : {
                            'other-key' : 'value',
                            'key' : 'value'
                        }
                    },
                    'other-key' : {
                        'key' : {
                            'other-key' : 'value',
                            'key' : 'value'
                        }
                    }
                },
                'other-key' : {
                    'key' : {
                        'key' : {
                            'other-key' : 'value',
                            'key' : 'value'
                        }
                    },
                    'other-key' : {
                        'key' : {
                            'other-key' : 'value',
                            'key' : {
                                'some-query-key' : 'value',
                                'key' : 'value'
                            }
                        }
                    }
                }
            }
        },
        'other-key' : 'value'
    }

    # Act
    queryTree = SettingHelper.querySetting('some-query-key',tree)

    # Assert
    assert {
        'key.key.some-query-key': {
            'key': {
                'key': {
                    'some-query-key': 'value',
                    'key': 'value'
                }
            },
            'other-key': {
                'key': {
                    'other-key': 'value',
                    'key': 'value'
                }
            }
        },
        'key.key.some-query-key.key.key.some-query-key': 'value',
        'key.key.other-key.other-key.key.key.some-query-key': 'value'
    } == queryTree

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS : True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **LOG_HELPER_SETTINGS
    }
)
def mustHandleSettingValueInFallbackSettingTree() :
    # Arrange
    settingFallbackFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource','fallback-application.yml'])
    settingFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource','referencing-fallback-application.yml'])
    FIRST_LONG_STRING = '''"""Hi
                every
            one
            """'''
    SECOND_LONG_STRING = '''"""Hi
                            every
                            one
                            this
                            is
                            the
                            deepest
                            long
                                        string
                            here
                            """'''
    THIRD_LONG_STRING = '''"""
                    me
                    being
        not
                    fshds
                    """'''

    # Act
    readdedSettingFallbackFilePath = SettingHelper.getSettingTree(settingFallbackFilePath)
    readdedSettingTree = SettingHelper.getSettingTree(settingFilePath, keepDepthInLongString=True, fallbackSettingTree=readdedSettingFallbackFilePath)
    # log.prettyPython(mustHandleSettingValueInFallbackSettingTree, 'readdedSettingTree', readdedSettingTree, logLevel=log.SETTING)

    # Assert
    assert [] == SettingHelper.getSetting('reffer-to.fallback-settings.empty.list', readdedSettingTree)
    assert 'ABCD -- [] -- EFGH' == SettingHelper.getSetting('reffer-to.fallback-settings.empty.list-in-between', readdedSettingTree)
    assert (()) == SettingHelper.getSetting('reffer-to.fallback-settings.empty.tuple', readdedSettingTree)
    assert 'ABCD -- () -- EFGH' == SettingHelper.getSetting('reffer-to.fallback-settings.empty.tuple-in-between', readdedSettingTree)
    assert {} == SettingHelper.getSetting('reffer-to.fallback-settings.empty.set-or-dictionary', readdedSettingTree)
    assert 'ABCD -- {} -- EFGH' == SettingHelper.getSetting('reffer-to.fallback-settings.empty.set-or-dictionary-in-between', readdedSettingTree)
    assert [
        'a',
        'b',
        'c'
    ] == SettingHelper.getSetting('reffer-to.fallback-settings.not-empty.list', readdedSettingTree)
    assert "ABCD -- ['a', 'b', 'c'] -- EFGH" == SettingHelper.getSetting('reffer-to.fallback-settings.not-empty.list-in-between', readdedSettingTree)
    assert (
        True,
        False,
        'None'
    ) == SettingHelper.getSetting('reffer-to.fallback-settings.not-empty.tuple', readdedSettingTree)
    assert "ABCD -- (True, False, 'None') -- EFGH" == SettingHelper.getSetting('reffer-to.fallback-settings.not-empty.tuple-in-between', readdedSettingTree)
    assert {} == SettingHelper.getSetting('reffer-to.fallback-settings.not-empty.set', readdedSettingTree)
    assert 'ABCD -- {} -- EFGH' == SettingHelper.getSetting('reffer-to.fallback-settings.not-empty.set-in-between', readdedSettingTree)
    assert {
        '1': 20,
        '2': 10,
        '3': 30
    } == SettingHelper.getSetting('reffer-to.fallback-settings.not-empty.dictionary', readdedSettingTree)
    assert "ABCD -- {'1': 20, '2': 10, '3': 30} -- EFGH" == SettingHelper.getSetting('reffer-to.fallback-settings.not-empty.dictionary-in-between', readdedSettingTree)
    assert "fallback value" == SettingHelper.getSetting('reffer-to.fallback-settings.string', readdedSettingTree)
    assert "ABCD -- fallback value -- EFGH" == SettingHelper.getSetting('reffer-to.fallback-settings.string-in-between', readdedSettingTree)
    assert 222233444 == SettingHelper.getSetting('reffer-to.fallback-settings.integer', readdedSettingTree)
    assert "ABCD -- 222233444 -- EFGH" == SettingHelper.getSetting('reffer-to.fallback-settings.integer-in-between', readdedSettingTree)
    assert 2.3 == SettingHelper.getSetting('reffer-to.fallback-settings.float', readdedSettingTree)
    assert "ABCD -- 2.3 -- EFGH" == SettingHelper.getSetting('reffer-to.fallback-settings.float-in-between', readdedSettingTree)
    assert True == SettingHelper.getSetting('reffer-to.fallback-settings.boolean', readdedSettingTree)
    assert "ABCD -- True -- EFGH" == SettingHelper.getSetting('reffer-to.fallback-settings.boolean-in-between', readdedSettingTree)
    assert 'None' == SettingHelper.getSetting('reffer-to.fallback-settings.none', readdedSettingTree)
    assert "ABCD -- None -- EFGH" == SettingHelper.getSetting('reffer-to.fallback-settings.none-in-between', readdedSettingTree)
    assert ObjectHelper.equals({
        'some-reference': {
            'much': {
                'before-its-assignment': 'delayed assignment value'
            },
            'before-its-assignment': 'delayed assignment value'
        },
        'other': {
            'root': {
                'key': 'other root value'
            }
        },
        'my': {
            'self-reference-key': 'self reference value',
            'other': {
                'self-reference-key': {
                    'as-well': 'other self reference value as well'
                },
                'repeated': {
                    'self-reference-key': {
                        'as-well': 'other repeated self reference value as well'
                    }
                }
            },
            'configuration-without-environment-variable-key': 'my default value',
            'configuration-without-environment-variable-key-with-value-surrounded-by-single-quotes': 'my default value',
            'configuration-without-environment-variable-key-and-space-after-colon': 'my default value',
            'configuration': 'self reference value',
            'own': {
                'configuration': 'self reference value',
                'very': {
                    'deep': {
                        'configuration': 'other root value'
                    }
                }
            },
            'other-with-other-name': {
                'self-reference-key': {
                    'as-well': 'other self reference value as well'
                },
                'configuration': 'self reference value',
                'configuration-as-well': 'other self reference value as well',
                'configuration-repeated-as-well': 'other repeated self reference value as well'
            },
            'override-case': {
                'overrider': 'overrider configuration'
            }
        },
        'long': {
            'string': FIRST_LONG_STRING
        },
        'deepest': {
            'long': {
                'string': {
                    'ever': {
                        'long': {
                            'string': SECOND_LONG_STRING
                        }
                    }
                }
            }
        },
        'not': {
            'idented': {
                'long': {
                    'string': THIRD_LONG_STRING
                }
            }
        },
        'new-key': 'new value',
        'my-list': {
            'numbers': [
                1,
                2,
                3,
                4
            ],
            'simple-strings': [
                'a',
                'b',
                'c',
                'd'
            ],
            'complex': [
                2,
                'b',
                'c',
                'd',
                1,
                2,
                True,
                True
            ],
            'with-elemets-surrounded-by-all-sorts-of-quotes': [
                'a',
                'b',
                'c',
                'd',
                'e',
                'f'
            ]
        },
        'specific-for': {
            'previous-assignment': 'delayed assignment value'
        },
        'some-key': {
            'with-an-enter-in-between-the-previous-one': f'{Constant.TRIPLE_SINGLE_QUOTE}  value  {Constant.TRIPLE_SINGLE_QUOTE} with spaces'
        },
        'it': {
            'contains': {
                'some-composed-key': {
                    'pointing-to': {
                        'a-late-value': 'abcd-- late value ----abcd---- late value ----abcd--efg',
                        'a-late-value-with-an-environment-variable-in-between': 'abcdit.contains.late-value--abcd--it.contains.late-value--abcd--efg'
                    }
                },
                'late-value': '-- late value --',
                'environment-variable': {
                    'only': {
                        'surrounded-by-default-values': 'ABCD --  -- EFGH',
                        'in-between-default-values': '''ABCD -- "some value followed by: "' and some following default value' -- EFGH'''
                    }
                },
                'refference': {
                    'to-a-late-definition': 'ABCD -- very late definiton value -- EFGH'
                },
                'one-setting-injection': 'abcdefg',
                'two-consecutive-setting-injection': 'abcdefghijklm',
                'one-inside-of-the-other-setting-injection': 'abcdefghijklm',
                'one-setting-injection-with-environment-variable': 'ABCDefgEFG',
                'one-inside-of-the-other-setting-injection-with-environment-variable': 'ABCDEFGEFGHIJKLMNOP',
                'two-consecutive-setting-injection-with-missing-environment-variable': 'abcdefghijklm'
            }
        },
        'very-late': {
            'definition': 'very late definiton value'
        },
        'handle': {
            'late': {
                'integer': 222233444,
                'float': 2.3,
                'boolean': True
            },
            'integer': 222233444,
            'float': 2.3,
            'boolean': True,
            'empty': {
                'list': [],
                'dictionary-or-set': {},
                'tuple': ()
            }
        },
        'some-not-string-selfreference': {
            'integer': 'ABCD -- 222233444 -- EFGH',
            'float': 'ABCD -- 2.3 -- EFGH',
            'boolean': 'ABCD -- True -- EFGH'
        },
        'reffer-to': {
            'fallback-settings': {
                'empty': {
                    'list': [],
                    'list-in-between': 'ABCD -- [] -- EFGH',
                    'tuple': (),
                    'tuple-in-between': 'ABCD -- () -- EFGH',
                    'set-or-dictionary': {},
                    'set-or-dictionary-in-between': 'ABCD -- {} -- EFGH'
                },
                'not-empty': {
                    'list': [
                        'a',
                        'b',
                        'c'
                    ],
                    'list-in-between': "ABCD -- ['a', 'b', 'c'] -- EFGH",
                    'tuple': (
                        True,
                        False,
                        'None'
                    ),
                    'tuple-in-between': "ABCD -- (True, False, 'None') -- EFGH",
                    'set': {},
                    'set-in-between': 'ABCD -- {} -- EFGH',
                    'dictionary': {
                        '1': 20,
                        '2': 10,
                        '3': 30
                    },
                    'dictionary-in-between': "ABCD -- {'1': 20, '2': 10, '3': 30} -- EFGH"
                },
                'string': 'fallback value',
                'string-in-between': 'ABCD -- fallback value -- EFGH',
                'integer': 222233444,
                'integer-in-between': 'ABCD -- 222233444 -- EFGH',
                'float': 2.3,
                'float-in-between': 'ABCD -- 2.3 -- EFGH',
                'boolean': True,
                'boolean-in-between': 'ABCD -- True -- EFGH',
                'none': 'None',
                'none-in-between': 'ABCD -- None -- EFGH'
            }
        },
        'some': {
            'key': 'value',
            'very': {
                'key': 'value',
                'similar': {
                    'key': 'value',
                    'tree': {
                        'key': 'value',
                        'with': {
                            'key': 'value',
                            'a': {
                                'key': 'value',
                                'slight': {
                                    'key': 'value',
                                    'difference': {
                                        'key': 'value',
                                        'right': {
                                            'here': 'anoter value'
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        },
        'print-status': True,
        'server': {
            'scheme': 'http',
            'port': 5001,
            'servlet': {
                'context-path': '/test-api'
            }
        },
        'swagger': {
            'schemes': [
                'http'
            ],
            'host': 'localhost:5001',
            'info': {
                'title': 'TestApi',
                'version': '0.0.1',
                'description': 'description',
                'terms-of-service': 'http://swagger.io/terms/',
                'contact': {
                    'name': 'Samuel Jansen',
                    'email': 'samuel.jansenn@gmail.com'
                },
                'license': {
                    'name': 'Apache 2.0 / MIT License',
                    'url': 'http://www.apache.org/licenses/LICENSE-2.0.html'
                }
            }
        },
        'fallback': {
            'string': 'fallback value',
            'integer': 222233444,
            'float': 2.3,
            'boolean': True,
            'none': 'None',
            'empty': {
                'list': [],
                'set-or-dictionary': {},
                'tuple': ()
            },
            'not-empty': {
                'list': [
                    'a',
                    'b',
                    'c'
                ],
                'dictionary': {
                    '1': 20,
                    '2': 10,
                    '3': 30
                },
                'set': {},
                'tuple': (
                    True,
                    False,
                    'None'
                )
            }
        },
        'flask-specific-port': 'flask run --host=0.0.0.0 --port=5001',
        'api': {
            'name': 'TestApi',
            'extension': 'yml',
            'dependency': {
                'update': False,
                'list': {
                    'web': [
                        'globals',
                        'python_helper',
                        'Popen',
                        'Path',
                        'numpy',
                        'pywin32',
                        'sqlalchemy'
                    ]
                }
            },
            'git': {
                'force-upgrade-command': 'pip install --upgrade --force python_framework'
            },
            'static-package': 'AppData\Local\Programs\Python\Python38-32\statics'
        }
    },
    readdedSettingTree, ignoreKeyList=['it'])

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS : True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **LOG_HELPER_SETTINGS
    }
)
def mustHandleSettingValueInFallbackSettingTree_whenFallbackSettingFilePathIsPassedInsted() :
    # Arrange
    settingFallbackFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource','fallback-application.yml'])
    settingFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource','referencing-fallback-application.yml'])
    FIRST_LONG_STRING = '''"""Hi
                every
            one
            """'''
    SECOND_LONG_STRING = '''"""Hi
                            every
                            one
                            this
                            is
                            the
                            deepest
                            long
                                        string
                            here
                            """'''
    THIRD_LONG_STRING = '''"""
                    me
                    being
        not
                    fshds
                    """'''

    # Act
    readdedSettingTree = SettingHelper.getSettingTree(settingFilePath, keepDepthInLongString=True, fallbackSettingFilePath=settingFallbackFilePath)
    # log.prettyPython(mustHandleSettingValueInFallbackSettingTree_whenFallbackSettingFilePathIsPassedInsted, 'readdedSettingTree', readdedSettingTree, logLevel=log.SETTING)

    # Assert
    assert [] == SettingHelper.getSetting('reffer-to.fallback-settings.empty.list', readdedSettingTree)
    assert 'ABCD -- [] -- EFGH' == SettingHelper.getSetting('reffer-to.fallback-settings.empty.list-in-between', readdedSettingTree)
    assert (()) == SettingHelper.getSetting('reffer-to.fallback-settings.empty.tuple', readdedSettingTree)
    assert 'ABCD -- () -- EFGH' == SettingHelper.getSetting('reffer-to.fallback-settings.empty.tuple-in-between', readdedSettingTree)
    assert {} == SettingHelper.getSetting('reffer-to.fallback-settings.empty.set-or-dictionary', readdedSettingTree)
    assert 'ABCD -- {} -- EFGH' == SettingHelper.getSetting('reffer-to.fallback-settings.empty.set-or-dictionary-in-between', readdedSettingTree)
    assert [
        'a',
        'b',
        'c'
    ] == SettingHelper.getSetting('reffer-to.fallback-settings.not-empty.list', readdedSettingTree)
    assert "ABCD -- ['a', 'b', 'c'] -- EFGH" == SettingHelper.getSetting('reffer-to.fallback-settings.not-empty.list-in-between', readdedSettingTree)
    assert (
        True,
        False,
        'None'
    ) == SettingHelper.getSetting('reffer-to.fallback-settings.not-empty.tuple', readdedSettingTree)
    assert "ABCD -- (True, False, 'None') -- EFGH" == SettingHelper.getSetting('reffer-to.fallback-settings.not-empty.tuple-in-between', readdedSettingTree)
    assert {} == SettingHelper.getSetting('reffer-to.fallback-settings.not-empty.set', readdedSettingTree)
    assert 'ABCD -- {} -- EFGH' == SettingHelper.getSetting('reffer-to.fallback-settings.not-empty.set-in-between', readdedSettingTree)
    assert {
        '1': 20,
        '2': 10,
        '3': 30
    } == SettingHelper.getSetting('reffer-to.fallback-settings.not-empty.dictionary', readdedSettingTree)
    assert "ABCD -- {'1': 20, '2': 10, '3': 30} -- EFGH" == SettingHelper.getSetting('reffer-to.fallback-settings.not-empty.dictionary-in-between', readdedSettingTree)
    assert "fallback value" == SettingHelper.getSetting('reffer-to.fallback-settings.string', readdedSettingTree)
    assert "ABCD -- fallback value -- EFGH" == SettingHelper.getSetting('reffer-to.fallback-settings.string-in-between', readdedSettingTree)
    assert 222233444 == SettingHelper.getSetting('reffer-to.fallback-settings.integer', readdedSettingTree)
    assert "ABCD -- 222233444 -- EFGH" == SettingHelper.getSetting('reffer-to.fallback-settings.integer-in-between', readdedSettingTree)
    assert 2.3 == SettingHelper.getSetting('reffer-to.fallback-settings.float', readdedSettingTree)
    assert "ABCD -- 2.3 -- EFGH" == SettingHelper.getSetting('reffer-to.fallback-settings.float-in-between', readdedSettingTree)
    assert True == SettingHelper.getSetting('reffer-to.fallback-settings.boolean', readdedSettingTree)
    assert "ABCD -- True -- EFGH" == SettingHelper.getSetting('reffer-to.fallback-settings.boolean-in-between', readdedSettingTree)
    assert 'None' == SettingHelper.getSetting('reffer-to.fallback-settings.none', readdedSettingTree)
    assert "ABCD -- None -- EFGH" == SettingHelper.getSetting('reffer-to.fallback-settings.none-in-between', readdedSettingTree)
    assert ObjectHelper.equals({
        'some-reference': {
            'much': {
                'before-its-assignment': 'delayed assignment value'
            },
            'before-its-assignment': 'delayed assignment value'
        },
        'other': {
            'root': {
                'key': 'other root value'
            }
        },
        'my': {
            'self-reference-key': 'self reference value',
            'other': {
                'self-reference-key': {
                    'as-well': 'other self reference value as well'
                },
                'repeated': {
                    'self-reference-key': {
                        'as-well': 'other repeated self reference value as well'
                    }
                }
            },
            'configuration-without-environment-variable-key': 'my default value',
            'configuration-without-environment-variable-key-with-value-surrounded-by-single-quotes': 'my default value',
            'configuration-without-environment-variable-key-and-space-after-colon': 'my default value',
            'configuration': 'self reference value',
            'own': {
                'configuration': 'self reference value',
                'very': {
                    'deep': {
                        'configuration': 'other root value'
                    }
                }
            },
            'other-with-other-name': {
                'self-reference-key': {
                    'as-well': 'other self reference value as well'
                },
                'configuration': 'self reference value',
                'configuration-as-well': 'other self reference value as well',
                'configuration-repeated-as-well': 'other repeated self reference value as well'
            },
            'override-case': {
                'overrider': 'overrider configuration'
            }
        },
        'long': {
            'string': FIRST_LONG_STRING
        },
        'deepest': {
            'long': {
                'string': {
                    'ever': {
                        'long': {
                            'string': SECOND_LONG_STRING
                        }
                    }
                }
            }
        },
        'not': {
            'idented': {
                'long': {
                    'string': THIRD_LONG_STRING
                }
            }
        },
        'new-key': 'new value',
        'my-list': {
            'numbers': [
                1,
                2,
                3,
                4
            ],
            'simple-strings': [
                'a',
                'b',
                'c',
                'd'
            ],
            'complex': [
                2,
                'b',
                'c',
                'd',
                1,
                2,
                True,
                True
            ],
            'with-elemets-surrounded-by-all-sorts-of-quotes': [
                'a',
                'b',
                'c',
                'd',
                'e',
                'f'
            ]
        },
        'specific-for': {
            'previous-assignment': 'delayed assignment value'
        },
        'some-key': {
            'with-an-enter-in-between-the-previous-one': '\'\'\'  value  \'\'\' with spaces'
        },
        'it': {
            'contains': {
                'some-composed-key': {
                    'pointing-to': {
                        'a-late-value': 'abcd-- late value ----abcd---- late value ----abcd--efg',
                        'a-late-value-with-an-environment-variable-in-between': 'abcdit.contains.late-value--abcd--it.contains.late-value--abcd--efg'
                    }
                },
                'late-value': '-- late value --',
                'environment-variable': {
                    'only': {
                        'surrounded-by-default-values': 'ABCD --  -- EFGH',
                        'in-between-default-values': '''ABCD -- "some value followed by: "\' and some following default value\' -- EFGH'''
                    }
                },
                'refference': {
                    'to-a-late-definition': 'ABCD -- very late definiton value -- EFGH'
                },
                'one-setting-injection': 'abcdefg',
                'two-consecutive-setting-injection': 'abcdefghijklm',
                'one-inside-of-the-other-setting-injection': 'abcdefghijklm',
                'one-setting-injection-with-environment-variable': 'ABCDefgEFG',
                'one-inside-of-the-other-setting-injection-with-environment-variable': 'ABCDEFGEFGHIJKLMNOP',
                'two-consecutive-setting-injection-with-missing-environment-variable': 'abcdefghijklm'
            }
        },
        'very-late': {
            'definition': 'very late definiton value'
        },
        'handle': {
            'late': {
                'integer': 222233444,
                'float': 2.3,
                'boolean': True
            },
            'integer': 222233444,
            'float': 2.3,
            'boolean': True,
            'empty': {
                'list': [],
                'dictionary-or-set': {},
                'tuple': ()
            }
        },
        'some-not-string-selfreference': {
            'integer': 'ABCD -- 222233444 -- EFGH',
            'float': 'ABCD -- 2.3 -- EFGH',
            'boolean': 'ABCD -- True -- EFGH'
        },
        'reffer-to': {
            'fallback-settings': {
                'empty': {
                    'list': [],
                    'list-in-between': 'ABCD -- [] -- EFGH',
                    'tuple': (),
                    'tuple-in-between': 'ABCD -- () -- EFGH',
                    'set-or-dictionary': {},
                    'set-or-dictionary-in-between': 'ABCD -- {} -- EFGH'
                },
                'not-empty': {
                    'list': [
                        'a',
                        'b',
                        'c'
                    ],
                    'list-in-between': "ABCD -- ['a', 'b', 'c'] -- EFGH",
                    'tuple': (
                        True,
                        False,
                        'None'
                    ),
                    'tuple-in-between': "ABCD -- (True, False, 'None') -- EFGH",
                    'set': {},
                    'set-in-between': 'ABCD -- {} -- EFGH',
                    'dictionary': {
                        '1': 20,
                        '2': 10,
                        '3': 30
                    },
                    'dictionary-in-between': "ABCD -- {'1': 20, '2': 10, '3': 30} -- EFGH"
                },
                'string': 'fallback value',
                'string-in-between': 'ABCD -- fallback value -- EFGH',
                'integer': 222233444,
                'integer-in-between': 'ABCD -- 222233444 -- EFGH',
                'float': 2.3,
                'float-in-between': 'ABCD -- 2.3 -- EFGH',
                'boolean': True,
                'boolean-in-between': 'ABCD -- True -- EFGH',
                'none': 'None',
                'none-in-between': 'ABCD -- None -- EFGH'
            }
        },
        'some': {
            'key': 'value',
            'very': {
                'key': 'value',
                'similar': {
                    'key': 'value',
                    'tree': {
                        'key': 'value',
                        'with': {
                            'key': 'value',
                            'a': {
                                'key': 'value',
                                'slight': {
                                    'key': 'value',
                                    'difference': {
                                        'key': 'value',
                                        'right': {
                                            'here': 'anoter value'
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        },
        'print-status': True,
        'server': {
            'scheme': 'http',
            'port': 5001,
            'servlet': {
                'context-path': '/test-api'
            }
        },
        'swagger': {
            'schemes': [
                'http'
            ],
            'host': 'localhost:5001',
            'info': {
                'title': 'TestApi',
                'version': '0.0.1',
                'description': 'description',
                'terms-of-service': 'http://swagger.io/terms/',
                'contact': {
                    'name': 'Samuel Jansen',
                    'email': 'samuel.jansenn@gmail.com'
                },
                'license': {
                    'name': 'Apache 2.0 / MIT License',
                    'url': 'http://www.apache.org/licenses/LICENSE-2.0.html'
                }
            }
        },
        'fallback': {
            'string': 'fallback value',
            'integer': 222233444,
            'float': 2.3,
            'boolean': True,
            'none': 'None',
            'empty': {
                'list': [],
                'set-or-dictionary': {},
                'tuple': ()
            },
            'not-empty': {
                'list': [
                    'a',
                    'b',
                    'c'
                ],
                'dictionary': {
                    '1': 20,
                    '2': 10,
                    '3': 30
                },
                'set': {},
                'tuple': (
                    True,
                    False,
                    'None'
                )
            }
        },
        'flask-specific-port': 'flask run --host=0.0.0.0 --port=5001',
        'api': {
            'name': 'TestApi',
            'extension': 'yml',
            'dependency': {
                'update': False,
                'list': {
                    'web': [
                        'globals',
                        'python_helper',
                        'Popen',
                        'Path',
                        'numpy',
                        'pywin32',
                        'sqlalchemy'
                    ]
                }
            },
            'git': {
                'force-upgrade-command': 'pip install --upgrade --force python_framework'
            },
            'static-package': 'AppData\Local\Programs\Python\Python38-32\statics'
        }
    },
    readdedSettingTree, ignoreKeyList=['it'])

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS : True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **LOG_HELPER_SETTINGS
    }
)
def getSettingTree_whenIsSettingKeyActuallyContainsSettingKey() :
    # Arrange
    settingFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource','framework.yml'])
    expected = {
        'database': {
            'dialect': 'a:b$c:d',
            'username': 'e:f?g:h',
            'password': 'i:j!k:l',
            'host': 'm:n*o:p',
            'port': '[q:r:s:t]',
            'schema': '(u:v:w:x)'
        },
        'environment': {
            'database': {
                'key': 'DATABASE_URL',
                'value': 'a:b$c:d://e:f?g:h:i:j!k:l@m:n*o:p:[q:r:s:t]/(u:v:w:x)'
            }
        },
        'server': {
            'scheme': 'https',
            'host': 'host',
            'servlet': {
                'context-path': '/test-api'
            },
            'port': 5050
        },
        'api': {
            'name': 'TestApi',
            'host-0': 'https://host',
            'host-1': 'https://host/test-api',
            'host-2': 'https://host:5050/test-api'
        },
        'swagger': {
            'host': 'host',
            'info': {
                'title': 'TestApi',
                'version': '0.0.1',
                'description': 'description',
                'terms-of-service': 'http://swagger.io/terms/',
                'contact': {
                    'name': 'Samuel Jansen',
                    'email': 'samuel.jansenn@gmail.com'
                },
                'license': {
                    'name': 'Apache 2.0 / MIT License',
                    'url': 'http://www.apache.org/licenses/LICENSE-2.0.html'
                }
            },
            'schemes': [
                'https'
            ]
        },
        'some': {
            'dictionary': {
                'yolo': 'yes',
                'another-yolo': 'no',
                'another-yolo-again': '',
                "{ 'again?'": "'yes' }"
            }
        }
    }

    # Act
    readdedSettingTree = SettingHelper.getSettingTree(settingFilePath)
    # log.prettyPython(getSettingTree_whenIsSettingKeyActuallyContainsSettingKey, 'readdedSettingTree', readdedSettingTree)

    # Assert
    assert "a:b$c:d://e:f?g:h:i:j!k:l@m:n*o:p:[q:r:s:t]/(u:v:w:x)" == SettingHelper.getSetting('environment.database.value', readdedSettingTree)
    assert ObjectHelper.equals(expected, readdedSettingTree)

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS : True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **LOG_HELPER_SETTINGS
    }
)
def getSettingTree_fallbackPriority() :
    # arrange
    defaultFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource','fallback-priority.yml'])
    localFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource','fallback-priority-local.yml'])
    expected = {
        'print-status': True,
        'server': {
            'scheme': 'http',
            'host': 'localhost',
            'host-and-port': 'localhost:5050',
            'port': 5050,
            'servlet': {
                'context-path': '/test-api'
            }
        },
        'has-it': {
            'or-not': '?',
            'here': '?'
        },
        'flask-specific-port': 'flask run --host=0.0.0.0 --port=5001',
        'api': {
            'name': 'TestApi',
            'extension': 'yml',
            'dependency': {
                'update': False,
                'list': {
                    'web': [
                        'globals',
                        'python_helper',
                        'Popen',
                        'Path',
                        'numpy',
                        'pywin32',
                        'sqlalchemy'
                    ]
                }
            },
            'git': {
                'force-upgrade-command': 'pip install --upgrade --force python_framework'
            },
            'static-package': 'AppData\Local\Programs\Python\Python38-32\statics',
            'list': []
        },
        'swagger': {
            'info': {
                'title': 'TestApi',
                'version': '0.0.1',
                'description': 'description',
                'terms-of-service': 'http://swagger.io/terms/',
                'contact': {
                    'name': 'Samuel Jansen',
                    'email': 'samuel.jansenn@gmail.com'
                },
                'license': {
                    'name': 'Apache 2.0 / MIT License',
                    'url': 'http://www.apache.org/licenses/LICENSE-2.0.html'
                }
            },
            'host': 'localhost:5050',
            'schemes': [
                'http'
            ]
        },
        'python': {
            'version': 3.9
        }
    }

    # act
    toAssert = SettingHelper.getSettingTree(localFilePath, keepDepthInLongString=True, fallbackSettingFilePath=defaultFilePath)
    # log.prettyPython(getSettingTree_fallbackPriority, 'localSettings', toAssert, logLevel=log.SETTING)

    assert ObjectHelper.equals(expected, toAssert), f'{StringHelper.prettyPython(expected)} == {StringHelper.prettyPython(toAssert)}'

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS : True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        'MY_COMPLEX_ENV' : ' -- my complex value -- ',
        'LATE_VALUE' : '-- late environment value --',
        'ONLY_ENVIRONMENT_VARIABLE' : 'only environment variable value',
        **LOG_HELPER_SETTINGS,
    }
)
def getSettingTree_otherApplication() :
    # arrange
    defaultFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource','other-application.yml'])
    localFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource','other-application-local.yml'])
    LOCAL_CONFIG_VALUE = 'local config setting value'
    FIRST_LONG_STRING = '''"""Hi
                every
            one
            """'''
    SECOND_LONG_STRING = '''"""Hi
                            every
                            one
                            this
                            is
                            the
                            deepest
                            long
                                        string
                            here
                            """'''
    THIRD_LONG_STRING = '''"""
                    me
                    being
        not
                    fshds
                    """'''
    expected = {
        'print-status': True,
        'local': {
            'config': {
                'setting-key': 'local config setting value'
            }
        },
        'database': {
            'dialect': 'a:b$c:d',
            'username': 'e:f?g:h',
            'password': 'i:j!k:l',
            'host': 'm:n*o:p',
            'port': '[q:r:s:t]',
            'schema': '(u:v:w:x)'
        },
        'environment': {
            'database': {
                'key': 'DATABASE_URL',
                'value': 'a:b$c:d://e:f?g:h:i:j!k:l@m:n*o:p:[q:r:s:t]/(u:v:w:x)'
            },
            'test': 'production',
            'missing': 'not at all'
        },
        'server': {
            'scheme': 'https',
            'host': 'host',
            'servlet': {
                'context-path': '/test-api'
            },
            'port': 5050
        },
        'api': {
            'host-0': 'https://host',
            'host-1': 'https://host/test-api',
            'host-2': 'https://host:5050/test-api',
            'name': 'Globals',
            'extension': 'yml',
            'dependency': {
                'update': False,
                'list': {
                    'web': [
                        'Popen',
                        'Path'
                    ],
                    'local': []
                }
            },
            'list': [
                'Globals'
            ],
            'language': 'EN-US',
            'git': {
                'url': 'https://github.com/SamuelJansen/',
                'extension': 'git'
            }
        },
        'swagger': {
            'host': 'host',
            'info': {
                'title': 'TestApi',
                'version': '0.0.1',
                'description': 'description',
                'terms-of-service': 'http://swagger.io/terms/',
                'contact': {
                    'name': 'Samuel Jansen',
                    'email': 'samuel.jansenn@gmail.com'
                },
                'license': {
                    'name': 'Apache 2.0 / MIT License',
                    'url': 'http://www.apache.org/licenses/LICENSE-2.0.html'
                }
            },
            'schemes': [
                'https'
            ]
        },
        'python': {
            'version': 3.9
        },
        'some-reference': {
            'much': {
                'before-its-assignment': 'delayed assignment value'
            },
            'before-its-assignment': 'delayed assignment value'
        },
        'other': {
            'root': {
                'key': 'other root value'
            }
        },
        'my': {
            'self-reference-key': 'self reference value',
            'other': {
                'self-reference-key': {
                    'as-well': 'other self reference value as well'
                },
                'repeated': {
                    'self-reference-key': {
                        'as-well': 'other repeated self reference value as well'
                    }
                }
            },
            'configuration-without-environment-variable-key': 'my default value',
            'configuration-without-environment-variable-key-with-value-surrounded-by-single-quotes': 'my default value',
            'configuration-without-environment-variable-key-and-space-after-colon': 'my default value',
            'own': {
                'very': {
                    'deep': {
                        'configuration': 'other root value'
                    }
                },
                'configuration': 'self reference value'
            },
            'other-with-other-name': {
                'self-reference-key': {
                    'as-well': 'other self reference value as well'
                },
                'configuration': 'self reference value',
                'configuration-as-well': 'other self reference value as well',
                'configuration-repeated-as-well': 'other repeated self reference value as well'
            },
            'override-case': {
                'overrider': 'overrider configuration'
            },
            'configuration': 'self reference value'
        },
        'long': {
            'string': FIRST_LONG_STRING
        },
        'deepest': {
            'long': {
                'string': {
                    'ever': {
                        'long': {
                            'string': SECOND_LONG_STRING
                        }
                    }
                }
            }
        },
        'not': {
            'idented': {
                'long': {
                    'string': THIRD_LONG_STRING
                }
            }
        },
        'new-key': 'new value',
        'my-list': {
            'numbers': [
                1,
                2,
                3,
                4
            ],
            'simple-strings': [
                'a',
                'b',
                'c',
                'd'
            ],
            'complex': [
                2,
                'b',
                'c',
                'd',
                1,
                2,
                True,
                True
            ],
            'with-elemets-surrounded-by-all-sorts-of-quotes': [
                'a',
                'b',
                'c',
                'd',
                'e',
                'f'
            ]
        },
        'specific-for': {
            'previous-assignment': 'delayed assignment value'
        },
        'some-key': {
            'with-an-enter-in-between-the-previous-one': "'''  value  ''' with spaces"
        },
        'it': {
            'contains': {
                'some-composed-key': {
                    'pointing-to': {
                        'a-late-value': 'abcd-- late value ----abcd---- late value ----abcd--efg',
                        'a-late-value-with-an-environment-variable-in-between': 'abcd-- late environment value ----abcd--it.contains.late-value--abcd--efg'
                    }
                },
                'late-value': '-- late value --',
                'environment-variable': {
                    'only': 'only environment variable value',
                    'surrounded-by-default-values': 'ABCD -- only environment variable value -- EFGH',
                    'in-between-default-values': """ABCD -- "some value followed by: "only environment variable value' and some following default value' -- EFGH"""
                },
                'refference': {
                    'to-a-late-definition': 'ABCD -- very late definiton value -- EFGH'
                },
                'one-setting-injection': 'abcdefg',
                'two-consecutive-setting-injection': 'abcdefghijklm',
                'one-inside-of-the-other-setting-injection': 'abcdefghijklm',
                'one-setting-injection-with-environment-variable': 'ABCD-- my complex value --EFG',
                'one-inside-of-the-other-setting-injection-with-environment-variable': 'ABCDEFGEFG-- my complex value --HIJKLMNOP',
                'two-consecutive-setting-injection-with-missing-environment-variable': 'abcdefghijklm'
            }
        },
        'very-late': {
            'definition': 'very late definiton value'
        },
        'handle': {
            'late': {
                'integer': 222233444,
                'float': 2.3,
                'boolean': True
            },
            'integer': 222233444,
            'float': 2.3,
            'boolean': True,
            'empty': {
                'list': [],
                'dictionary-or-set': {},
                'tuple': ()
            }
        },
        'some': {
            'dictionary': {
                'yolo': 'yes',
                'another-yolo': 'no',
                'another-yolo-again': '',
                f'''{'{'}{" 'again?'"}''': f'''{"'yes' "}{'}'}'''
            }
        },
        'some-not-string-selfreference': {
            'integer': 'ABCD -- 222233444 -- EFGH',
            'float': 'ABCD -- 2.3 -- EFGH',
            'boolean': 'ABCD -- True -- EFGH'
        }
    }

    # act
    # log.prettyPython(getSettingTree_fallbackPriority, 'expected', expected, logLevel=log.SETTING)
    falback = SettingHelper.getSettingTree(defaultFilePath, keepDepthInLongString=True)
    # log.prettyPython(getSettingTree_fallbackPriority, 'falback', falback, logLevel=log.SETTING)
    toAssert = SettingHelper.getSettingTree(localFilePath, keepDepthInLongString=True, fallbackSettingFilePath=defaultFilePath, fallbackSettingTree=falback)
    # log.prettyPython(getSettingTree_fallbackPriority, 'toAssert', toAssert, logLevel=log.SETTING)

    # assert
    assert LOCAL_CONFIG_VALUE == SettingHelper.getSetting('local.config.setting-key', toAssert)
    assert True == SettingHelper.getSetting('print-status', toAssert)
    assert 'Globals' == SettingHelper.getSetting('api.name', toAssert)
    assert "a:b$c:d://e:f?g:h:i:j!k:l@m:n*o:p:[q:r:s:t]/(u:v:w:x)" == SettingHelper.getSetting('environment.database.value', toAssert), SettingHelper.getSetting('environment.database.value', toAssert)
    assert expected['long']['string'] == toAssert['long']['string']
    assert expected['deepest']['long']['string']['ever']['long']['string'] == toAssert['deepest']['long']['string']['ever']['long']['string']
    assert expected['not']['idented']['long']['string'] == toAssert['not']['idented']['long']['string']
    assert ObjectHelper.equals(expected['some']['dictionary'], toAssert['some']['dictionary'])
    assert ObjectHelper.equals(expected, toAssert)

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS : True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        'MIRACLE' : 'yes.this.is.possible',
        **LOG_HELPER_SETTINGS,
    }
)
def getSettingTree_whenThereAreNoneValuesAllOverThePlace() :
    TEST_AMOUNT = 16
    for index in range(TEST_AMOUNT) :
        # arrange
        noneValuesFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource', 'nonevalues', f'none-values{index}.yml'])
        exception = None

        # act
        try :
            toAssert = SettingHelper.getSettingTree(noneValuesFilePath)
            # log.prettyPython(getSettingTree_whenThereAreNoneValuesAllOverThePlace, f'{index}', toAssert, logLevel=log.DEBUG)
        except Exception as e :
            exception = e
            # print(f'{index}: {exception}')

        # assert
        assert ObjectHelper.isNotNone(exception)
        assert StringHelper.isNotBlank(str(exception))

    SECOND_TEST_AMOUNT = 31
    for index in range(TEST_AMOUNT, SECOND_TEST_AMOUNT) :
        # arrange
        noneValuesFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource', 'nonevalues', f'none-values{index}.yml'])
        exception = None

        # act
        try :
            toAssert = SettingHelper.getSettingTree(noneValuesFilePath)
            # log.prettyPython(getSettingTree_whenThereAreNoneValuesAllOverThePlace, f'{index}', toAssert, logLevel=log.DEBUG)
        except Exception as e :
            exception = e
            # print(f'{index}: {exception}')

        # assert
        assert ObjectHelper.isNone(exception)

    ENVIRONMENT_INJECTION = 7
    for index in range(ENVIRONMENT_INJECTION) :
        # arrange
        falbackFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource', 'nonevalues', f'import-from{index}.yml'])
        noneValuesFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource', 'nonevalues', f'the-one-to-import{index}.yml'])
        exception = None

        fallback = None

        assert 'yes.this.is.possible' == EnvironmentHelper.get('MIRACLE')

        # act
        try :
            fallback = SettingHelper.getSettingTree(falbackFilePath)
            toAssert = SettingHelper.getSettingTree(noneValuesFilePath, fallbackSettingFilePath=falbackFilePath, fallbackSettingTree=fallback)
            # log.prettyPython(getSettingTree_whenThereAreNoneValuesAllOverThePlace, f'{index}', toAssert, logLevel=log.DEBUG)
        except Exception as e :
            exception = e
            # print(fallback)
            print(f'{index}: {exception}')

        # assert
        assert ObjectHelper.isNone(exception)

    FAIL_ENVIRONMENT_INJECTION = 7
    for index in range(FAIL_ENVIRONMENT_INJECTION) :
        # arrange
        failFalbackFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource', 'nonevalues', f'fail-import-from{index}.yml'])
        failNoneValuesFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource', 'nonevalues', f'fail-the-one-to-import{index}.yml'])
        exception = None

        fallback = None

        # act
        try :
            failFallback = SettingHelper.getSettingTree(failFalbackFilePath)
            toAssert = SettingHelper.getSettingTree(failNoneValuesFilePath, fallbackSettingFilePath=failNoneValuesFilePath, fallbackSettingTree=failFallback)
            # log.prettyPython(getSettingTree_whenThereAreNoneValuesAllOverThePlace, f'{index}', toAssert, logLevel=log.DEBUG)
        except Exception as e :
            exception = e
            # print(fallback)
            # print(f'{index}: {exception}')

        # assert
        assert ObjectHelper.isNotNone(exception)
        assert StringHelper.isNotBlank(str(exception))

@Test(
    environmentVariables={
        'ENVIRONMENT_BOOLEAN_VALUE': True,
        log.ENABLE_LOGS_WITH_COLORS : True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **LOG_HELPER_SETTINGS,
    }
)
def getBooleanSetting():
    # arrange
    defaultFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource','boolean-application.yml'])
    localFilePath = str(EnvironmentHelper.OS_SEPARATOR).join(['python_helper', 'api', 'test', 'api', 'resource','boolean-application-local.yml'])

    # act
    falseValue = SettingHelper.getSettingTree(defaultFilePath, keepDepthInLongString=True)['boolean']['value']
    trueValue = SettingHelper.getSettingTree(localFilePath, keepDepthInLongString=True, fallbackSettingFilePath=defaultFilePath, fallbackSettingTree=SettingHelper.getSettingTree(defaultFilePath, keepDepthInLongString=True))['boolean']['value']
    environmentInjectionFalseValue = SettingHelper.getSettingTree(defaultFilePath, keepDepthInLongString=True)['boolean']['environment-injection']['value']
    environmentInjectionTrueValue = SettingHelper.getSettingTree(localFilePath, keepDepthInLongString=True, fallbackSettingFilePath=defaultFilePath, fallbackSettingTree=SettingHelper.getSettingTree(defaultFilePath, keepDepthInLongString=True))['boolean']['environment-injection']['value']
    isTrue = SettingHelper.getSettingTree(localFilePath, keepDepthInLongString=True, fallbackSettingFilePath=defaultFilePath, fallbackSettingTree=SettingHelper.getSettingTree(defaultFilePath, keepDepthInLongString=True))['boolean']['environment-injection']['is-true']
    shouldBeFalse = SettingHelper.getSettingTree(defaultFilePath, keepDepthInLongString=True)['boolean']['environment-injection']['this-is-true']
    shouldBeTrue = SettingHelper.getSettingTree(localFilePath, keepDepthInLongString=True, fallbackSettingFilePath=defaultFilePath, fallbackSettingTree=SettingHelper.getSettingTree(defaultFilePath, keepDepthInLongString=True))['boolean']['environment-injection']['this-is-true']

    thisIsAlsoTruePartial = SettingHelper.getSettingTree(localFilePath, keepDepthInLongString=True, fallbackSettingFilePath=defaultFilePath, fallbackSettingTree=SettingHelper.getSettingTree(defaultFilePath, keepDepthInLongString=True))
    thisIsAlsoFalsePartial = SettingHelper.getSettingTree(localFilePath, keepDepthInLongString=True, fallbackSettingFilePath=defaultFilePath, fallbackSettingTree=SettingHelper.getSettingTree(defaultFilePath, keepDepthInLongString=True))
    # print(thisIsAlsoTruePartial)
    # print(thisIsAlsoFalsePartial)

    thisIsAlsoTrue = thisIsAlsoTruePartial['boolean']['environment-injection']['this-is-also-true']
    thisIsAlsoFalse = thisIsAlsoFalsePartial['boolean']['environment-injection']['this-is-also-false']

    #assert
    assert trueValue, f'trueValue should be True with type bool, but is {trueValue} with type {type(trueValue)}'
    assert bool == type(trueValue), f'trueValue should be True with type bool, but is {trueValue} with type {type(trueValue)}'
    assert not falseValue, f'falseValue should be False with type bool, but is {falseValue} with type {type(falseValue)}'
    assert bool == type(falseValue), f'falseValue should be False with type bool, but is {falseValue} with type {type(falseValue)}'

    assert environmentInjectionTrueValue, f'environmentInjectionTrueValue should be True with type bool, but is {environmentInjectionTrueValue} with type {type(environmentInjectionTrueValue)}'
    assert bool == type(environmentInjectionTrueValue), f'environmentInjectionTrueValue should be True with type bool, but is {environmentInjectionTrueValue} with type {type(environmentInjectionTrueValue)}'
    assert not environmentInjectionFalseValue, f'environmentInjectionFalseValue should be False with type bool, but is {environmentInjectionFalseValue} with type {type(environmentInjectionFalseValue)}'
    assert bool == type(environmentInjectionFalseValue), f'environmentInjectionFalseValue should be False with type bool, but is {environmentInjectionFalseValue} with type {type(environmentInjectionFalseValue)}'

    assert isTrue, f'isTrue with type bool, but is {isTrue} with type {type(isTrue)}'
    assert bool == type(isTrue), f'isTrue with type bool, but is {isTrue} with type {type(isTrue)}'
    assert not shouldBeFalse, f'shouldBeFalse with type bool, but is {shouldBeFalse} with type {type(shouldBeFalse)}'
    assert bool == type(shouldBeFalse), f'shouldBeFalse with type bool, but is {shouldBeFalse} with type {type(shouldBeFalse)}'
    assert shouldBeTrue, f'shouldBeTrue with type bool, but is {shouldBeTrue} with type {type(shouldBeTrue)}'
    assert bool == type(shouldBeTrue), f'shouldBeTrue with type bool, but is {shouldBeTrue} with type {type(shouldBeTrue)}'

    assert thisIsAlsoTrue, f'thisIsAlsoTrue with type bool, but is {thisIsAlsoTrue} with type {type(thisIsAlsoTrue)}'
    assert bool == type(thisIsAlsoTrue), f'thisIsAlsoTrue with type bool, but is {thisIsAlsoTrue} with type {type(thisIsAlsoTrue)}'
    assert not thisIsAlsoFalse, f'thisIsAlsoFalse with type bool, but is {thisIsAlsoFalse} with type {type(thisIsAlsoFalse)}'
    assert bool == type(thisIsAlsoFalse), f'thisIsAlsoFalse with type bool, but is {thisIsAlsoFalse} with type {type(thisIsAlsoFalse)}'



import datetime
from python_helper.api.src.domain import Constant as c
from python_helper.api.src.service import ObjectHelper, StringHelper, RandomHelper, LogHelper

DEFAULT_DATETIME_PATTERN = '%Y-%m-%d %H:%M:%S'
DEFAULT_DATE_PATTERN = '%Y-%m-%d'
DEFAULT_TIME_PATTERN = '%H:%M:%S'

DATETIME_FULL_PATTERN = '%Y-%m-%d %H:%M:%S.%f'
TIME_FULL_PATTERN = '%H:%M:%S.%f'

PATTERNS = (
    DEFAULT_DATETIME_PATTERN,
    DEFAULT_DATE_PATTERN,
    DEFAULT_TIME_PATTERN,
    DATETIME_FULL_PATTERN,
    TIME_FULL_PATTERN
)

DATETIME_PATTERNS = (
    DEFAULT_DATETIME_PATTERN,
    DATETIME_FULL_PATTERN
)

DATE_PATTERNS = (
    DEFAULT_DATE_PATTERN
)

TIME_PATTERNS = (
    DEFAULT_TIME_PATTERN,
    TIME_FULL_PATTERN
)

DEFAULT_TIME_BEGIN = '00:00:00'
DEFAULT_TIME_END = '23:59:59'


def toString(givenDatetime, pattern=DEFAULT_DATETIME_PATTERN):
    return givenDatetime if ObjectHelper.isNone(givenDatetime) or isinstance(givenDatetime, str) else parseToString(givenDatetime, pattern=pattern)

def parseToString(given, pattern=DEFAULT_DATETIME_PATTERN):
    # return str(parseToPattern(str(given), pattern=pattern))
    # return str(given)
    return str(forcedlyParse(str(given), pattern=pattern))

def parseToPattern(given, pattern=DEFAULT_DATETIME_PATTERN, timedelta=False):
    given = given.strip()
    if StringHelper.isNotBlank(given):
        parsed = datetime.datetime.strptime(given, pattern)
        if timedelta and pattern in TIME_PATTERNS :
            return timeDelta(hours=parsed.hour, minutes=parsed.minute, seconds=parsed.second, milliseconds=parsed.millisecond, microseconds=parsed.microsecond)
        if pattern in DATETIME_PATTERNS :
            return parsed
        elif pattern in DATE_PATTERNS :
            return parsed.date()
        elif pattern in TIME_PATTERNS :
            return parsed.time()

def forcedlyParse(given, pattern=DEFAULT_DATETIME_PATTERN, timedelta=False):
    parsed = None
    for pattern in [pattern, *PATTERNS] :
        try :
            parsed = parseToPattern(given, pattern=pattern, timedelta=timedelta)
            break
        except Exception as exception :
            pass
    return parsed

def parseToDateTime(givenDatetime, pattern=DEFAULT_DATETIME_PATTERN):
    return givenDatetime if ObjectHelper.isNone(givenDatetime) or not isinstance(givenDatetime, str) else parseToPattern(givenDatetime, pattern=pattern)

def forcedlyGetDateTime(givenDatetime, pattern=DEFAULT_DATETIME_PATTERN):
    return givenDatetime if ObjectHelper.isNone(givenDatetime) or not isinstance(givenDatetime, str) else forcedlyParse(givenDatetime, pattern=pattern)

def forcedlyGetDate(givenDate, pattern=DEFAULT_DATE_PATTERN):
    return givenDate if ObjectHelper.isNone(givenDate) or not isinstance(givenDate, str) else forcedlyParse(givenDate, pattern=pattern)

def forcedlyGetTime(givenTime, pattern=DEFAULT_TIME_PATTERN):
    return givenTime if ObjectHelper.isNone(givenTime) or not isinstance(givenTime, str) else forcedlyParse(givenTime, pattern=pattern)

def forcedlyGetInterval(givenTime, pattern=DEFAULT_DATETIME_PATTERN):
    return givenTime if ObjectHelper.isNone(givenTime) or not isinstance(givenTime, str) else forcedlyParse(givenTime, pattern=pattern, timedelta=True)

def timeDelta(days=0, hours=0, minutes=0, seconds=0, milliseconds=0, microseconds=0, **kwargs):
    return datetime.timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds, milliseconds=milliseconds, microseconds=microseconds, **kwargs)

def dateTimeNow():
    # return datetime.datetime.utcnow()
    return datetime.datetime.now()

def now():
    return dateTimeNow()

def dateOf(dateTime=None, pattern=DEFAULT_DATE_PATTERN):
    return dateTime.date()

def timeOf(dateTime=None, pattern=DEFAULT_TIME_PATTERN):
    return dateTime.time()

def dateNow():
    # datetime.date.today()
    return dateOf(dateTime=dateTimeNow())

def timeNow():
    return timeOf(dateTime=dateTimeNow())

def timestampNow():
    return datetime.datetime.timestamp(dateTimeNow())

def ofTimestamp(timestamp):
    # return datetime.datetime.utcfromtimestamp(timestamp)
    return datetime.datetime.fromtimestamp(timestamp)

def of(dateTime=None, date=None, time=None, pattern=DEFAULT_DATETIME_PATTERN):
    if isinstance(dateTime, datetime.datetime):
        return forcedlyParse(str(dateTime), pattern=pattern)
    datePattern = None
    timePattern = None
    if ObjectHelper.isNotNone(pattern):
        splittedPattern = pattern.split()
        if 1 < len(splittedPattern):
            datePattern = splittedPattern[0]
            timePattern = splittedPattern[1]
    if ObjectHelper.isNone(datePattern) or ObjectHelper.isNone(timePattern):
        datePattern = DEFAULT_DATE_PATTERN
        timePattern = DEFAULT_TIME_PATTERN
    if ObjectHelper.isNotNone(dateTime):
        if isinstance(dateTime, str):
            date = dateOf(dateTime, pattern=datePattern).split()[0]
            time = timeOf(dateTime, pattern=timePattern).split()[1]
        else:
            date = dateOf(dateTime, pattern=datePattern)
            time = timeOf(dateTime, pattern=timePattern)
    return datetime.datetime.combine(
        forcedlyGetDate(date if ObjectHelper.isNotNone(date) else dateNow(), pattern=datePattern),
        forcedlyGetTime(time if ObjectHelper.isNotNone(time) else DEFAULT_TIME_BEGIN, pattern=timePattern)
    )

def timestampOf(dateTime=None, date=None, time=None, pattern=DATETIME_FULL_PATTERN):
    return datetime.datetime.timestamp(of(date=date, time=time, dateTime=dateTime, pattern=pattern))

def plusSeconds(givenDateTimeOrTime, seconds=None, deltaInSeconds=None):
    if ObjectHelper.isNotNone(seconds):
        deltaInMinutes = timeDelta(seconds=seconds)
    if isinstance(givenDateTimeOrTime, datetime.time):
        givenDateTimeOrTime = forcedlyParse(f'{str(dateNow())} {givenDateTimeOrTime}')
    return forcedlyGetDateTime(str(givenDateTimeOrTime)) + deltaInMinutes

def minusSeconds(givenDateTimeOrTime, seconds=None, deltaInSeconds=None):
    if ObjectHelper.isNotNone(seconds):
        deltaInMinutes = timeDelta(seconds=seconds)
    if isinstance(givenDateTimeOrTime, datetime.time):
        givenDateTimeOrTime = forcedlyParse(f'{str(dateNow())} {givenDateTimeOrTime}')
    return forcedlyGetDateTime(str(givenDateTimeOrTime)) - deltaInMinutes

def plusMinutes(givenDateTimeOrTime, minutes=None, deltaInMinutes=None):
    if ObjectHelper.isNotNone(minutes):
        deltaInMinutes = timeDelta(seconds=minutes*60)
    if isinstance(givenDateTimeOrTime, datetime.time):
        givenDateTimeOrTime = forcedlyParse(f'{str(dateNow())} {givenDateTimeOrTime}')
    return forcedlyGetDateTime(str(givenDateTimeOrTime)) + deltaInMinutes

def minusMinutes(givenDateTimeOrTime, minutes=None, deltaInMinutes=None):
    if ObjectHelper.isNotNone(minutes):
        deltaInMinutes = timeDelta(minutes=minutes)
    if isinstance(givenDateTimeOrTime, datetime.time):
        givenDateTimeOrTime = forcedlyParse(f'{str(dateNow())} {givenDateTimeOrTime}')
    return forcedlyGetDateTime(str(givenDateTimeOrTime)) - deltaInMinutes

def plusDays(givenDay, days=None, deltaInDays=None):
    if ObjectHelper.isNotNone(days):
        deltaInDays = timeDelta(days=days)
    return forcedlyGetDateTime(str(givenDay)) + deltaInDays

def minusDays(givenDay, days=None, deltaInDays=None):
    if ObjectHelper.isNotNone(minutes):
        deltaInDays = timeDelta(days=days)
    return forcedlyGetDateTime(str(givenDay)) - deltaInDays

def getDefaultTimeBegin():
    return forcedlyGetTime(DEFAULT_TIME_BEGIN)

def getDatetimeMonthBegin():
    return parseToPattern(c.SPACE.join([c.DASH.join(str(dateTimeNow()).split()[0].split(c.DASH)[:-1] + ['01']), DEFAULT_TIME_BEGIN]))

def getTodayDateAndTodayTime():
    dateTime = dateTimeNow()
    return dateOf(dateTime=dateTime), timeOf(dateTime=dateTime)

def getTodayDateTimeBegin():
    return parseToDateTime(f'{dateNow()} {DEFAULT_TIME_BEGIN}')

def getTodayDateTimeEnd():
    return parseToDateTime(f'{dateNow()} {DEFAULT_TIME_END}')

###- deprecated
def getWeekDay(dateTime=None, date=None, time=None):
    if ObjectHelper.isNotNone(dateTime):
        return forcedlyGetDateTime(dateTime).weekday()
    elif ObjectHelper.isNotNone(date) and ObjectHelper.isNotNone(time):
        return of(date=forcedlyGetDate(date), time=forcedlyGetTime(time)).weekday()
    elif ObjectHelper.isNotNone(date):
        return of(date=forcedlyGetDate(date), time=forcedlyGetTime(DEFAULT_TIME_END)).weekday()
    return dateTimeNow().weekday()

def getWeekDayOf(dateTime=None, date=None, time=None):
    '''Repetition of getWeekDay, just for semantic'''
    return getWeekDay(dateTime=dateTime, date=date, time=time)

def addNoise(givenDatetime):
    return givenDatetime + timeDelta(milliseconds=RandomHelper.integer(0,999))



import string as randstr
from random import randint
import random
from python_helper.api.src.domain import Constant as c
from python_helper.api.src.helper import RandomHelperHelper
from python_helper.api.src.service import ObjectHelper

DEFAULT_MINIMUM_LENGHT = 0
DEFAULT_MAXIMUM_LENGHT = 10

def string(minimum=DEFAULT_MINIMUM_LENGHT, maximum=DEFAULT_MAXIMUM_LENGHT) :
    maximum = RandomHelperHelper.updateMaximum(minimum, maximum)
    stringRange = integer(minimum=minimum, maximum=maximum)
    return c.NOTHING.join(random.choice(randstr.ascii_uppercase + randstr.digits) for _ in range(stringRange))

def integer(minimum=DEFAULT_MINIMUM_LENGHT, maximum=DEFAULT_MAXIMUM_LENGHT) :
    maximum = RandomHelperHelper.updateMaximum(minimum, maximum)
    return randint(minimum, maximum)

def float(minimum=DEFAULT_MINIMUM_LENGHT, maximum=DEFAULT_MAXIMUM_LENGHT) :
    maximum = RandomHelperHelper.updateMaximum(minimum, maximum)
    difference = maximum - minimum
    return minimum + difference * random.random()

def sample(collection, length=None) :
    if ObjectHelper.isCollection(collection) :
        if ObjectHelper.isDictionary(collection) :
            if ObjectHelper.isNone(length) :
                key = RandomHelperHelper.sampleCollection(list(collection.keys()), list)
                return { key : collection[key] }
            sampleDictionaryKeys = RandomHelperHelper.sampleCollection(list(collection.keys()), list, length=length)
            sampleDictionary = {}
            for key in sampleDictionaryKeys :
                sampleDictionary[key] = collection[key]
            return sampleDictionary
        elif ObjectHelper.isList(collection) :
            return RandomHelperHelper.sampleCollection(collection, list, length=length)
        elif ObjectHelper.isSet(collection) :
            return RandomHelperHelper.sampleCollection(collection, set, length=length)
        elif ObjectHelper.isTuple(collection) :
            return RandomHelperHelper.sampleCollection(collection, tuple, length=length)
    raise Exception(f'The "{collection}" argument is not a collection')



import json
from python_helper import ObjectHelper, StringHelper, SettingHelper, Constant, log, Test, ReflectionHelper, RandomHelper

# LOG_HELPER_SETTINGS = {
#     log.LOG : True,
#     log.SUCCESS : True,
#     log.SETTING : True,
#     log.DEBUG : True,
#     log.WARNING : True,
#     log.FAILURE : True,
#     log.WRAPPER : True,
#     log.ERROR : True,
    # log.TEST : False,
    # log.ENABLE_LOGS_WITH_COLORS : True,
#     SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT
# }

LOG_HELPER_SETTINGS = {
    log.LOG : False,
    log.SUCCESS : False,
    log.SETTING : False,
    log.DEBUG : False,
    log.WARNING : False,
    log.FAILURE : False,
    log.WRAPPER : False,
    log.ERROR : False,
    log.TEST : False
}

class MyDto:
    def __init__(self, myAttribute, myOther, myThirdList) :
        self.myAttribute = myAttribute
        self.myOther = myOther
        self.myThirdList = myThirdList

class MyOtherDto:
    def __init__(self, myAttribute) :
        self.myAttribute = myAttribute

class MyThirdDto :
    def __init__(self, my, myAttribute) :
        self.my = my
        self.myAttribute = myAttribute

DICTIONARY_INSTANCE = {
    'my_none_value' : None,
    'my_none_value-as_string' : 'None',
    'string_key_as_string' : 'value',
    'integer_key_as_string' : 12345678901,
    10987654321 : 12345678901,
    'bool_key_as_string' : True,
    False : True,
    'float_key_as_string' : 12345.678901,
    109876.54321 : 12345.678901,
    'list_key_as_string' : [
        'my',
        'list',
        'elements'
    ],
    'list_of_list_key_as_string' : [
        [
            'my',
            'first',
            'list',
            False,
            12345.678901
        ],
        [
            'my',
            'second',
            'list',
            True,
            10987654321
        ]
    ],
    'set_of_string_key_as_string' : {
        'a',
        'True',
        '2',
        '3.4',
        True,
        2,
        2.4
    }
}
JSON_INSTANCE = json.loads(StringHelper.prettyJson(DICTIONARY_INSTANCE))

TEST_SETTINGS = {}

@Test(
    environmentVariables={**{}, **LOG_HELPER_SETTINGS},
    **TEST_SETTINGS
)
def basicMethods() :
    # arrange
    def generatorInstance() :
        while True :
            yield False
            break
    STR_INSTANCE = str()
    BOOLEAN_INSTANCE = bool()
    INTEGER_INSTANCE = int()
    FLOAT_INSTANCE = float()
    DICTIONARY_INSTANCE = dict()
    LIST_INSTANCE = list()
    TUPLE_INSTANCE = tuple()
    SET_INSTANCE = set()
    GENERATOR_INSTANCE = generatorInstance()

    STR_FILLED_INSTANCE = 'str()'
    BOOLEAN_FILLED_INSTANCE = True
    INTEGER_FILLED_INSTANCE = 2
    FLOAT_FILLED_INSTANCE = 3.3
    DICTIONARY_FILLED_INSTANCE = {'dict()':dict()}
    LIST_FILLED_INSTANCE = [list(),list()]
    TUPLE_FILLED_INSTANCE = (tuple(),tuple())
    SET_FILLED_INSTANCE = {'set()',2}

    # act

    # assert
    assert ObjectHelper.isNotNone(STR_INSTANCE)
    assert ObjectHelper.isNotNone(BOOLEAN_INSTANCE)
    assert ObjectHelper.isNotNone(INTEGER_INSTANCE)
    assert ObjectHelper.isNotNone(FLOAT_INSTANCE)
    assert ObjectHelper.isNotNone(DICTIONARY_INSTANCE)
    assert ObjectHelper.isNotNone(LIST_INSTANCE)
    assert ObjectHelper.isNotNone(TUPLE_INSTANCE)
    assert ObjectHelper.isNotNone(SET_INSTANCE)
    assert ObjectHelper.isNotNone(GENERATOR_INSTANCE)

    assert ObjectHelper.isNotNone(STR_FILLED_INSTANCE)
    assert ObjectHelper.isNotNone(BOOLEAN_FILLED_INSTANCE)
    assert ObjectHelper.isNotNone(INTEGER_FILLED_INSTANCE)
    assert ObjectHelper.isNotNone(FLOAT_FILLED_INSTANCE)
    assert ObjectHelper.isNotNone(DICTIONARY_FILLED_INSTANCE)
    assert ObjectHelper.isNotNone(LIST_FILLED_INSTANCE)
    assert ObjectHelper.isNotNone(TUPLE_FILLED_INSTANCE)
    assert ObjectHelper.isNotNone(SET_FILLED_INSTANCE)

    assert not ObjectHelper.isNone(STR_FILLED_INSTANCE)
    assert not ObjectHelper.isNone(BOOLEAN_FILLED_INSTANCE)
    assert not ObjectHelper.isNone(INTEGER_FILLED_INSTANCE)
    assert not ObjectHelper.isNone(FLOAT_FILLED_INSTANCE)
    assert not ObjectHelper.isNone(DICTIONARY_FILLED_INSTANCE)
    assert not ObjectHelper.isNone(LIST_FILLED_INSTANCE)
    assert not ObjectHelper.isNone(TUPLE_FILLED_INSTANCE)
    assert not ObjectHelper.isNone(SET_FILLED_INSTANCE)
    assert not ObjectHelper.isNone(GENERATOR_INSTANCE)

    assert not ObjectHelper.isList(STR_FILLED_INSTANCE)
    assert not ObjectHelper.isList(BOOLEAN_FILLED_INSTANCE)
    assert not ObjectHelper.isList(INTEGER_FILLED_INSTANCE)
    assert not ObjectHelper.isList(FLOAT_FILLED_INSTANCE)
    assert not ObjectHelper.isList(DICTIONARY_FILLED_INSTANCE)
    assert ObjectHelper.isList(LIST_FILLED_INSTANCE)
    assert not ObjectHelper.isList(TUPLE_FILLED_INSTANCE)
    assert not ObjectHelper.isList(SET_FILLED_INSTANCE)
    assert not ObjectHelper.isList(GENERATOR_INSTANCE)

    assert ObjectHelper.isNotList(STR_FILLED_INSTANCE)
    assert ObjectHelper.isNotList(BOOLEAN_FILLED_INSTANCE)
    assert ObjectHelper.isNotList(INTEGER_FILLED_INSTANCE)
    assert ObjectHelper.isNotList(FLOAT_FILLED_INSTANCE)
    assert ObjectHelper.isNotList(DICTIONARY_FILLED_INSTANCE)
    assert not ObjectHelper.isNotList(LIST_FILLED_INSTANCE)
    assert ObjectHelper.isNotList(TUPLE_FILLED_INSTANCE)
    assert ObjectHelper.isNotList(SET_FILLED_INSTANCE)
    assert ObjectHelper.isNotList(GENERATOR_INSTANCE)

    assert not ObjectHelper.isSet(STR_FILLED_INSTANCE)
    assert not ObjectHelper.isSet(BOOLEAN_FILLED_INSTANCE)
    assert not ObjectHelper.isSet(INTEGER_FILLED_INSTANCE)
    assert not ObjectHelper.isSet(FLOAT_FILLED_INSTANCE)
    assert not ObjectHelper.isSet(DICTIONARY_FILLED_INSTANCE)
    assert not ObjectHelper.isSet(LIST_FILLED_INSTANCE)
    assert not ObjectHelper.isSet(TUPLE_FILLED_INSTANCE)
    assert ObjectHelper.isSet(SET_FILLED_INSTANCE)
    assert not ObjectHelper.isSet(GENERATOR_INSTANCE)

    assert ObjectHelper.isNotSet(STR_FILLED_INSTANCE)
    assert ObjectHelper.isNotSet(BOOLEAN_FILLED_INSTANCE)
    assert ObjectHelper.isNotSet(INTEGER_FILLED_INSTANCE)
    assert ObjectHelper.isNotSet(FLOAT_FILLED_INSTANCE)
    assert ObjectHelper.isNotSet(DICTIONARY_FILLED_INSTANCE)
    assert ObjectHelper.isNotSet(LIST_FILLED_INSTANCE)
    assert ObjectHelper.isNotSet(TUPLE_FILLED_INSTANCE)
    assert not ObjectHelper.isNotSet(SET_FILLED_INSTANCE)
    assert ObjectHelper.isNotSet(GENERATOR_INSTANCE)

    assert not ObjectHelper.isTuple(STR_FILLED_INSTANCE)
    assert not ObjectHelper.isTuple(BOOLEAN_FILLED_INSTANCE)
    assert not ObjectHelper.isTuple(INTEGER_FILLED_INSTANCE)
    assert not ObjectHelper.isTuple(FLOAT_FILLED_INSTANCE)
    assert not ObjectHelper.isTuple(DICTIONARY_FILLED_INSTANCE)
    assert not ObjectHelper.isTuple(LIST_FILLED_INSTANCE)
    assert ObjectHelper.isTuple(TUPLE_FILLED_INSTANCE)
    assert not ObjectHelper.isTuple(SET_FILLED_INSTANCE)
    assert not ObjectHelper.isTuple(GENERATOR_INSTANCE)

    assert ObjectHelper.isNotTuple(STR_FILLED_INSTANCE)
    assert ObjectHelper.isNotTuple(BOOLEAN_FILLED_INSTANCE)
    assert ObjectHelper.isNotTuple(INTEGER_FILLED_INSTANCE)
    assert ObjectHelper.isNotTuple(FLOAT_FILLED_INSTANCE)
    assert ObjectHelper.isNotTuple(DICTIONARY_FILLED_INSTANCE)
    assert ObjectHelper.isNotTuple(LIST_FILLED_INSTANCE)
    assert not ObjectHelper.isNotTuple(TUPLE_FILLED_INSTANCE)
    assert ObjectHelper.isNotTuple(SET_FILLED_INSTANCE)
    assert ObjectHelper.isNotTuple(GENERATOR_INSTANCE)

    assert not ObjectHelper.isDictionary(STR_FILLED_INSTANCE)
    assert not ObjectHelper.isDictionary(BOOLEAN_FILLED_INSTANCE)
    assert not ObjectHelper.isDictionary(INTEGER_FILLED_INSTANCE)
    assert not ObjectHelper.isDictionary(FLOAT_FILLED_INSTANCE)
    assert ObjectHelper.isDictionary(DICTIONARY_FILLED_INSTANCE)
    assert not ObjectHelper.isDictionary(LIST_FILLED_INSTANCE)
    assert not ObjectHelper.isDictionary(TUPLE_FILLED_INSTANCE)
    assert not ObjectHelper.isDictionary(SET_FILLED_INSTANCE)
    assert not ObjectHelper.isDictionary(GENERATOR_INSTANCE)

    assert ObjectHelper.isNotDictionary(STR_FILLED_INSTANCE)
    assert ObjectHelper.isNotDictionary(BOOLEAN_FILLED_INSTANCE)
    assert ObjectHelper.isNotDictionary(INTEGER_FILLED_INSTANCE)
    assert ObjectHelper.isNotDictionary(FLOAT_FILLED_INSTANCE)
    assert not ObjectHelper.isNotDictionary(DICTIONARY_FILLED_INSTANCE)
    assert ObjectHelper.isNotDictionary(LIST_FILLED_INSTANCE)
    assert ObjectHelper.isNotDictionary(TUPLE_FILLED_INSTANCE)
    assert ObjectHelper.isNotDictionary(SET_FILLED_INSTANCE)
    assert ObjectHelper.isNotDictionary(GENERATOR_INSTANCE)

    assert not ObjectHelper.isCollection(STR_FILLED_INSTANCE)
    assert not ObjectHelper.isCollection(BOOLEAN_FILLED_INSTANCE)
    assert not ObjectHelper.isCollection(INTEGER_FILLED_INSTANCE)
    assert not ObjectHelper.isCollection(FLOAT_FILLED_INSTANCE)
    assert ObjectHelper.isCollection(DICTIONARY_FILLED_INSTANCE)
    assert ObjectHelper.isCollection(LIST_FILLED_INSTANCE)
    assert ObjectHelper.isCollection(TUPLE_FILLED_INSTANCE)
    assert ObjectHelper.isCollection(SET_FILLED_INSTANCE)
    assert not ObjectHelper.isCollection(GENERATOR_INSTANCE)

    assert ObjectHelper.isNotCollection(STR_FILLED_INSTANCE)
    assert ObjectHelper.isNotCollection(BOOLEAN_FILLED_INSTANCE)
    assert ObjectHelper.isNotCollection(INTEGER_FILLED_INSTANCE)
    assert ObjectHelper.isNotCollection(FLOAT_FILLED_INSTANCE)
    assert not ObjectHelper.isNotCollection(DICTIONARY_FILLED_INSTANCE)
    assert not ObjectHelper.isNotCollection(LIST_FILLED_INSTANCE)
    assert not ObjectHelper.isNotCollection(TUPLE_FILLED_INSTANCE)
    assert not ObjectHelper.isNotCollection(SET_FILLED_INSTANCE)
    assert ObjectHelper.isNotCollection(GENERATOR_INSTANCE)

    assert not ObjectHelper.isDictionaryClass(STR_FILLED_INSTANCE)
    assert not ObjectHelper.isDictionaryClass(BOOLEAN_FILLED_INSTANCE)
    assert not ObjectHelper.isDictionaryClass(INTEGER_FILLED_INSTANCE)
    assert not ObjectHelper.isDictionaryClass(FLOAT_FILLED_INSTANCE)
    assert not ObjectHelper.isDictionaryClass(DICTIONARY_FILLED_INSTANCE)
    assert not ObjectHelper.isDictionaryClass(LIST_FILLED_INSTANCE)
    assert not ObjectHelper.isDictionaryClass(TUPLE_FILLED_INSTANCE)
    assert not ObjectHelper.isDictionaryClass(SET_FILLED_INSTANCE)
    assert not ObjectHelper.isDictionaryClass(GENERATOR_INSTANCE)

    assert ObjectHelper.isNotDictionaryClass(STR_FILLED_INSTANCE)
    assert ObjectHelper.isNotDictionaryClass(BOOLEAN_FILLED_INSTANCE)
    assert ObjectHelper.isNotDictionaryClass(INTEGER_FILLED_INSTANCE)
    assert ObjectHelper.isNotDictionaryClass(FLOAT_FILLED_INSTANCE)
    assert ObjectHelper.isNotDictionaryClass(DICTIONARY_FILLED_INSTANCE)
    assert ObjectHelper.isNotDictionaryClass(LIST_FILLED_INSTANCE)
    assert ObjectHelper.isNotDictionaryClass(TUPLE_FILLED_INSTANCE)
    assert ObjectHelper.isNotDictionaryClass(SET_FILLED_INSTANCE)
    assert ObjectHelper.isNotDictionaryClass(GENERATOR_INSTANCE)

    assert not ObjectHelper.isDictionaryClass(type(STR_FILLED_INSTANCE))
    assert not ObjectHelper.isDictionaryClass(type(BOOLEAN_FILLED_INSTANCE))
    assert not ObjectHelper.isDictionaryClass(type(INTEGER_FILLED_INSTANCE))
    assert not ObjectHelper.isDictionaryClass(type(FLOAT_FILLED_INSTANCE))
    assert ObjectHelper.isDictionaryClass(type(DICTIONARY_FILLED_INSTANCE))
    assert not ObjectHelper.isDictionaryClass(type(LIST_FILLED_INSTANCE))
    assert not ObjectHelper.isDictionaryClass(type(TUPLE_FILLED_INSTANCE))
    assert not ObjectHelper.isDictionaryClass(type(SET_FILLED_INSTANCE))
    assert not ObjectHelper.isDictionaryClass(type(GENERATOR_INSTANCE))

    assert ObjectHelper.isNotDictionaryClass(type(STR_FILLED_INSTANCE))
    assert ObjectHelper.isNotDictionaryClass(type(BOOLEAN_FILLED_INSTANCE))
    assert ObjectHelper.isNotDictionaryClass(type(INTEGER_FILLED_INSTANCE))
    assert ObjectHelper.isNotDictionaryClass(type(FLOAT_FILLED_INSTANCE))
    assert not ObjectHelper.isNotDictionaryClass(type(DICTIONARY_FILLED_INSTANCE))
    assert ObjectHelper.isNotDictionaryClass(type(LIST_FILLED_INSTANCE))
    assert ObjectHelper.isNotDictionaryClass(type(TUPLE_FILLED_INSTANCE))
    assert ObjectHelper.isNotDictionaryClass(type(SET_FILLED_INSTANCE))
    assert ObjectHelper.isNotDictionaryClass(type(GENERATOR_INSTANCE))

    assert ObjectHelper.isNativeClass(type(STR_FILLED_INSTANCE))
    assert ObjectHelper.isNativeClass(type(BOOLEAN_FILLED_INSTANCE))
    assert ObjectHelper.isNativeClass(type(INTEGER_FILLED_INSTANCE))
    assert ObjectHelper.isNativeClass(type(FLOAT_FILLED_INSTANCE))
    assert not ObjectHelper.isNativeClass(type(DICTIONARY_FILLED_INSTANCE))
    assert not ObjectHelper.isNativeClass(type(LIST_FILLED_INSTANCE))
    assert not ObjectHelper.isNativeClass(type(TUPLE_FILLED_INSTANCE))
    assert not ObjectHelper.isNativeClass(type(SET_FILLED_INSTANCE))
    assert ObjectHelper.isNativeClass(type(GENERATOR_INSTANCE))

    assert not ObjectHelper.isNotNativeClass(type(STR_FILLED_INSTANCE))
    assert not ObjectHelper.isNotNativeClass(type(BOOLEAN_FILLED_INSTANCE))
    assert not ObjectHelper.isNotNativeClass(type(INTEGER_FILLED_INSTANCE))
    assert not ObjectHelper.isNotNativeClass(type(FLOAT_FILLED_INSTANCE))
    assert ObjectHelper.isNotNativeClass(type(DICTIONARY_FILLED_INSTANCE))
    assert ObjectHelper.isNotNativeClass(type(LIST_FILLED_INSTANCE))
    assert ObjectHelper.isNotNativeClass(type(TUPLE_FILLED_INSTANCE))
    assert ObjectHelper.isNotNativeClass(type(SET_FILLED_INSTANCE))
    assert not ObjectHelper.isNotNativeClass(type(GENERATOR_INSTANCE))

    assert ObjectHelper.isNativeClassInstance(STR_FILLED_INSTANCE)
    assert ObjectHelper.isNativeClassInstance(BOOLEAN_FILLED_INSTANCE)
    assert ObjectHelper.isNativeClassInstance(INTEGER_FILLED_INSTANCE)
    assert ObjectHelper.isNativeClassInstance(FLOAT_FILLED_INSTANCE)
    assert not ObjectHelper.isNativeClassInstance(DICTIONARY_FILLED_INSTANCE)
    assert not ObjectHelper.isNativeClassInstance(LIST_FILLED_INSTANCE)
    assert not ObjectHelper.isNativeClassInstance(TUPLE_FILLED_INSTANCE)
    assert not ObjectHelper.isNativeClassInstance(SET_FILLED_INSTANCE)
    assert ObjectHelper.isNativeClassInstance(GENERATOR_INSTANCE)

    assert not ObjectHelper.isNotNativeClassIsntance(STR_FILLED_INSTANCE)
    assert not ObjectHelper.isNotNativeClassIsntance(BOOLEAN_FILLED_INSTANCE)
    assert not ObjectHelper.isNotNativeClassIsntance(INTEGER_FILLED_INSTANCE)
    assert not ObjectHelper.isNotNativeClassIsntance(FLOAT_FILLED_INSTANCE)
    assert ObjectHelper.isNotNativeClassIsntance(DICTIONARY_FILLED_INSTANCE)
    assert ObjectHelper.isNotNativeClassIsntance(LIST_FILLED_INSTANCE)
    assert ObjectHelper.isNotNativeClassIsntance(TUPLE_FILLED_INSTANCE)
    assert ObjectHelper.isNotNativeClassIsntance(SET_FILLED_INSTANCE)
    assert not ObjectHelper.isNotNativeClassIsntance(GENERATOR_INSTANCE)

    assert ObjectHelper.isNone(None)
    assert not ObjectHelper.isNotNone(None)
    assert not ObjectHelper.isList(None)
    assert ObjectHelper.isNotList(None)
    assert not ObjectHelper.isSet(None)
    assert ObjectHelper.isNotSet(None)
    assert not ObjectHelper.isTuple(None)
    assert ObjectHelper.isNotTuple(None)
    assert not ObjectHelper.isDictionary(None)
    assert ObjectHelper.isNotDictionary(None)
    assert not ObjectHelper.isCollection(None)
    assert ObjectHelper.isNotCollection(None)
    assert not ObjectHelper.isDictionaryClass(None)
    assert ObjectHelper.isNotDictionaryClass(None)
    assert not ObjectHelper.isNativeClass(None)
    assert ObjectHelper.isNotNativeClass(None)
    assert not ObjectHelper.isNativeClassInstance(None)
    assert ObjectHelper.isNotNativeClassIsntance(None)

    assert not ObjectHelper.isNone(type(None))
    assert ObjectHelper.isNotNone(type(None))
    assert not ObjectHelper.isList(type(None))
    assert ObjectHelper.isNotList(type(None))
    assert not ObjectHelper.isDictionary(type(None))
    assert ObjectHelper.isNotDictionary(type(None))
    assert not ObjectHelper.isCollection(type(None))
    assert ObjectHelper.isNotCollection(type(None))
    assert not ObjectHelper.isDictionaryClass(type(None))
    assert ObjectHelper.isNotDictionaryClass(type(None))
    assert not ObjectHelper.isNativeClass(type(None))
    assert ObjectHelper.isNotNativeClass(type(None))
    assert not ObjectHelper.isNativeClassInstance(type(None))
    assert ObjectHelper.isNotNativeClassIsntance(type(None))

@Test(
    environmentVariables={**{}, **LOG_HELPER_SETTINGS},
    **TEST_SETTINGS
)
def mustAssertEquals() :
    # arrange
    dictionaryInstance = {**{},**JSON_INSTANCE}
    someDictionary = {
        'a' : 'b',
        'c' : 'd',
        'e' : {
            'f' : 'g',
            't' : [
                {
                    's1',
                    's1',
                    's3',
                    1,
                    3.3,
                    False,
                    None
                },
                {
                    's1',
                    False,
                    's3',
                    3.3,
                    1,
                    None
                }
            ],
            1 : 7,
            False : 2.3
        },
        2.2 : {
            False,
            2,
            None,
            'string'
        },
        'tuple' : (
            2,
            3,
            '3',
            9
        )
    }
    someDictionaryList = [
        someDictionary,
        someDictionary
    ]
    someOtherDictionary = {
        'c' : 'd',
        'a' : 'b',
        'e' : {
            False : 2.3,
            1 : 7,
            'f' : 'g',
            't' : [
                {
                    3.3,
                    's1',
                    's3',
                    1,
                    None,
                    False
                },
                {
                    's1',
                    's1',
                    's3',
                    None,
                    1,
                    3.3,
                    False
                }
            ]
        },
        2.2 : {
            False,
            2,
            None,
            'string'
        },
        'tuple' : (
            2,
            9,
            3,
            '3'
        )
    }
    someOtherDictionaryList = [
        someOtherDictionary,
        someOtherDictionary
    ]
    differentDictionary = {
        'c' : 'd',
        'a' : 'b',
        'e' : {
            False : 2.3,
            1 : 7,
            'f' : 'g',
            't' : [
                {
                    's1',
                    's3',
                    1,
                    3.3,
                    False,
                    None
                },
                {
                    's1',
                    's1',
                    's3',
                    1,
                    3.3,
                    False,
                    None
                }
            ]
        },
        2.2 : {
            False,
            2,
            None,
            str()
        },
        'tuple' : (
            9,
            2,
            3,
            '3'
        )
    }
    differentDictionaryList = [
        differentDictionary,
        differentDictionary
    ]
    aList = [
        {
            'beginAtDate': '2021-03-11',
            'beginAtDatetime': '2021-03-11 08:30:00',
            'beginAtTime': '08:30:00',
            'endAtDate': '2021-03-11',
            'endAtDatetime': '2021-03-11 08:30:00',
            'endAtTime': '08:30:00',
            'id': None,
            'intervalTime': '2021-03-11 08:30:00',
            'timedelta': '08:30:00'
        },
        {
            'beginAtDate': '2021-03-11',
            'beginAtDatetime': '2021-03-11 08:30:00',
            'beginAtTime': '08:30:00',
            'endAtDate': '2021-03-11',
            'endAtDatetime': '2021-03-11 08:30:00',
            'endAtTime': '08:30:00',
            'id': None,
            'intervalTime': '2021-03-11 08:30:00',
            'timedelta': '08:30:00'
        }
    ]
    bList = [
        {
            'beginAtDate': '2021-03-11',
            'beginAtDatetime': '2021-03-11 08:30:00',
            'beginAtTime': '08:30:00',
            'endAtDate': '2021-03-11',
            'endAtDatetime': '2021-03-11 08:30:00',
            'endAtTime': '08:30:00',
            'id': None,
            'intervalTime': '2021-03-11 08:30:00',
            'timedelta': '8:30:00'
        },
        {
            'beginAtDate': '2021-03-11',
            'beginAtDatetime': '2021-03-11 08:30:00',
            'beginAtTime': '08:30:00',
            'endAtDate': '2021-03-11',
            'endAtDatetime': '2021-03-11 08:30:00',
            'endAtTime': '08:30:00',
            'id': None,
            'intervalTime': '2021-03-11 08:30:00',
            'timedelta': '8:30:00'
        }
    ]

    # act
    toAssert = ObjectHelper.equals(dictionaryInstance, JSON_INSTANCE, ignoreCharactereList=[Constant.NEW_LINE])
    unsortedDictionaryToAssert = ObjectHelper.equals(someDictionary, someOtherDictionary)
    unsortedDictionaryListToAssert = ObjectHelper.equals(someDictionaryList, someOtherDictionaryList)
    notEqualsToAssert = ObjectHelper.equals(someDictionary, differentDictionary)
    notEqualsListToAssert = ObjectHelper.equals(someDictionaryList, differentDictionaryList)

    # assert
    assert toAssert
    assert unsortedDictionaryToAssert
    assert not notEqualsToAssert
    assert unsortedDictionaryListToAssert
    assert not notEqualsListToAssert
    assert not ObjectHelper.equals(aList, bList)
    assert ObjectHelper.equals(aList, bList, ignoreKeyList = ['timedelta'])

@Test(
    environmentVariables={**{}, **LOG_HELPER_SETTINGS},
    **TEST_SETTINGS
)
def mustIgnoreKeyCorrectly() :
    # arrange
    expected = {**{},**DICTIONARY_INSTANCE}
    anotherDictionaryInstance = {**{},**DICTIONARY_INSTANCE}
    IGNORABLE_KEY = 'ignorableKey'
    anotherDictionaryInstance[IGNORABLE_KEY] = 'ignorableValue'

    # act
    toAssert = ObjectHelper.filterIgnoreKeyList(anotherDictionaryInstance,[IGNORABLE_KEY])

    # assert
    assert ObjectHelper.equals(expected, toAssert)

@Test()
def equal_whenListOfDictionaries() :
    # arrange
    null = 'null'
    LIST_OF_DICTIONARIES = [
        {
            "myAttribute": "NW2",
            "myOther": {
                "myAttribute": "34PDZB"
            },
            "myThirdList": [
                {
                    "my": {
                        "myAttribute": "X1HC",
                        "myOther": {
                            "myAttribute": "34PDZB"
                        },
                        "myThirdList": null
                    },
                    "myAttribute": 9
                }
            ]
        },
        {
            "myAttribute": "",
            "myOther": null,
            "myThirdList": [
                {
                    "my": {
                        "myAttribute": "U",
                        "myOther": null,
                        "myThirdList": null
                    },
                    "myAttribute": 3
                }
            ]
        },
        {
            "myAttribute": "HNQ7QKW2",
            "myOther": {
                "myAttribute": "V9OXKD8"
            },
            "myThirdList": [
                {
                    "my": {
                        "myAttribute": "PVYA",
                        "myOther": {
                            "myAttribute": "V9OXKD8"
                        },
                        "myThirdList": null
                    },
                    "myAttribute": 10
                }
            ]
        }
    ]
    DIFFERENT_LIST_OF_DICTIONARIES = [
        {
            "myAttribute": "NW2",
            "myOther": {
                "myAttribute": "34PDZB"
            },
            "myThirdList": [
                {
                    "my": {
                        "myAttribute": "X1HC",
                        "myOther": {
                            "myAttribute": RandomHelper.integer()
                        },
                        "myThirdList": null
                    },
                    "myAttribute": 9
                }
            ]
        },
        {
            "myAttribute": "",
            "myOther": null,
            "myThirdList": [
                {
                    "my": {
                        "myAttribute": "U",
                        "myOther": null,
                        "myThirdList": null
                    },
                    "myAttribute": 3
                }
            ]
        },
        {
            "myAttribute": "HNQ7QKW2",
            "myOther": {
                "myAttribute": "V9OXKD8"
            },
            "myThirdList": [
                {
                    "my": {
                        "myAttribute": "PVYA",
                        "myOther": {
                            "myAttribute": "V9OXKD8"
                        },
                        "myThirdList": null
                    },
                    "myAttribute": 10
                }
            ]
        }
    ]

    # act
    # assert
    assert ObjectHelper.equals(LIST_OF_DICTIONARIES, [
        {
            "myAttribute": "NW2",
            "myOther": {
                "myAttribute": "34PDZB"
            },
            "myThirdList": [
                {
                    "my": {
                        "myAttribute": "X1HC",
                        "myOther": {
                            "myAttribute": "34PDZB"
                        },
                        "myThirdList": null
                    },
                    "myAttribute": 9
                }
            ]
        },
        {
            "myAttribute": "",
            "myOther": null,
            "myThirdList": [
                {
                    "my": {
                        "myAttribute": "U",
                        "myOther": null,
                        "myThirdList": null
                    },
                    "myAttribute": 3
                }
            ]
        },
        {
            "myAttribute": "HNQ7QKW2",
            "myOther": {
                "myAttribute": "V9OXKD8"
            },
            "myThirdList": [
                {
                    "my": {
                        "myAttribute": "PVYA",
                        "myOther": {
                            "myAttribute": "V9OXKD8"
                        },
                        "myThirdList": null
                    },
                    "myAttribute": 10
                }
            ]
        }
    ])
    assert False == ObjectHelper.equals(LIST_OF_DICTIONARIES, [{}, {}, {}])
    assert False == ObjectHelper.equals(LIST_OF_DICTIONARIES, DIFFERENT_LIST_OF_DICTIONARIES)

@Test()
def equal_whenObjects() :
    # arrange
    a = RandomHelper.string()
    b = RandomHelper.string()
    c = RandomHelper.string()
    otherA = MyOtherDto(RandomHelper.string())
    otherB = MyOtherDto(RandomHelper.string())
    otherC = MyOtherDto(RandomHelper.string())
    myFirst = MyDto(None, None, None)
    mySecond = MyDto(None, None, None)
    myThird = MyDto(None, None, None)
    thirdOne = RandomHelper.integer()
    thirdTwo = RandomHelper.integer()
    thirdThree = RandomHelper.integer()
    myThirdOne = [MyThirdDto(myFirst, thirdOne)]
    myThirdTwo = [MyThirdDto(mySecond, thirdTwo)]
    myThirdThree = [MyThirdDto(myThird, thirdThree)]
    expected = [MyDto(a, otherA, myThirdOne), MyDto(b, otherB, myThirdTwo), MyDto(c, otherC, myThirdThree)]
    null = 'null'
    inspectEquals = False

    # act
    toAssert = [MyDto(a, otherA, myThirdOne), MyDto(b, otherB, myThirdTwo), MyDto(c, otherC, myThirdThree)]
    another = [MyDto(a, otherA, [MyThirdDto(myFirst, thirdOne)]), MyDto(b, otherB, myThirdTwo), MyDto(c, otherC, myThirdThree)]
    another[0].myThirdList[0].my = MyDto(
        MyDto(None, None, None),
        expected[0].myThirdList[0].my.myOther,
        expected[0].myThirdList[0].my.myThirdList
    )

    # assert
    assert False == (expected == toAssert), f'False == ({expected} == {toAssert}): {False == (expected == toAssert)}'
    assert ObjectHelper.equals(expected, toAssert)
    assert ObjectHelper.equals(toAssert, expected)
    assert ObjectHelper.isNotNone(expected[0].myThirdList[0].my), expected[0].myThirdList[0].my
    assert expected[0].myThirdList[0].my == toAssert[0].myThirdList[0].my
    assert ObjectHelper.equals(expected[0].myThirdList[0].my, toAssert[0].myThirdList[0].my)
    assert ObjectHelper.isNone(expected[0].myThirdList[0].my.myThirdList)
    assert ObjectHelper.equals(expected[0].myThirdList[0].my.myThirdList, toAssert[0].myThirdList[0].my.myThirdList)
    assert ObjectHelper.equals(expected[1].myThirdList[0], toAssert[1].myThirdList[0])
    assert ObjectHelper.equals(toAssert[1].myThirdList[0], expected[1].myThirdList[0])
    assert False == (expected == another), f'False == ({expected} == {another}): False == {(expected == another)}'
    assert False == ObjectHelper.equals(expected, another, muteLogs=not inspectEquals)
    assert False == ObjectHelper.equals(another, expected, muteLogs=not inspectEquals)
    assert False == ObjectHelper.equals(another, toAssert, muteLogs=not inspectEquals)
    assert False == ObjectHelper.equals(toAssert, another, muteLogs=not inspectEquals)
    assert False == ObjectHelper.equals(expected, [MyDto(None, None, None), MyDto(None, None, None), MyDto(None, None, None)])

@Test()
def equal_whenDictionary() :
    #arrange
    firstDict = {'b': 'c', 'a': 'd'}
    secondDict = {'a': 'd', 'b': 'c'}

    #act
    firstDictSorted = ObjectHelper.sortIt(firstDict)
    secondDictSorted = ObjectHelper.sortIt(secondDict)

    #assert
    assert firstDictSorted == secondDictSorted
    # print(f'{firstDictSorted} == {secondDictSorted}: {firstDictSorted == secondDictSorted}')

    # a = RandomHelper.string()
    # b = RandomHelper.string()
    # c = RandomHelper.string()
    # otherA = MyOtherDto(RandomHelper.string())
    # otherB = MyOtherDto(RandomHelper.string())
    # otherC = MyOtherDto(RandomHelper.string())
    # myFirst = MyDto(RandomHelper.string(), otherA, None)
    # mySecond = MyDto(RandomHelper.string(), otherB, None)
    # myThird = MyDto(RandomHelper.string(), otherC, None)
    # thirdOne = RandomHelper.integer()
    # thirdTwo = RandomHelper.integer()
    # thirdThree = RandomHelper.integer()
    # myThirdOne = MyThirdDto(myFirst, thirdOne)
    # myThirdTwo = MyThirdDto(mySecond, thirdTwo)
    # myThirdThree = MyThirdDto(myThird, thirdThree)
    # expected = [MyDto(a, otherA, myThirdOne), MyDto(b, otherB, myThirdTwo), MyDto(c, otherC, myThirdThree)]
    #
    # # act
    # toAssert = [MyDto(a, otherA, myThirdOne), MyDto(b, otherB, myThirdTwo), MyDto(c, otherC, myThirdThree)]
    #
    # # assert
    # assert False == (expected == toAssert)
    # assert ObjectHelper.equals(expected, toAssert)
    # assert False == ObjectHelper.equals(expected, [MyDto(None, None, None), MyDto(None, None, None), MyDto(None, None, None)])



from python_helper.api.src.domain import Constant as c
from python_helper.api.src.service import ObjectHelper
from python_helper.api.src.helper import StringHelperHelper

PAST = 'past'
PRESENT = 'present'
FUTURE = 'future'

def isBlank(thing) :
    return isinstance(thing, str) and c.BLANK == thing

def isNotBlank(thing) :
    return isinstance(thing, str) and not c.BLANK == thing

def filterJson(json, extraCharacterList=None) :
    characterList = [c.NEW_LINE,c.BAR_N]
    if isinstance(extraCharacterList, list) :
        characterList += extraCharacterList
    filteredJson = json
    for character in characterList :
        filteredJson = removeCharactere(character,filteredJson)
    return filteredJson.replace(c.SYSTEM_TAB,c.TAB)

def removeCharactere(character,string) :
    if isNotBlank(character) and isNotBlank(string) :
        filteredString = c.BLANK.join(string.strip().split(character))
        return filteredString.replace(character,c.BLANK)
    return string

def join(stringList, character=c.BLANK):
    return character.join(stringList)

def prettyPython(
        outterValue,
        quote = c.SINGLE_QUOTE,
        tabCount = 0,
        nullValue = c.NONE,
        trueValue = c.TRUE,
        falseValue = c.FALSE,
        withColors = False,
        joinAtReturn = True
    ) :
    '''It always sort sets'''
    if ObjectHelper.isCollection(outterValue) :
        if isinstance(outterValue, list) :
            collectionType = c.TYPE_LIST
        elif isinstance(outterValue, set) :
            collectionType = c.TYPE_SET
        elif isinstance(outterValue, tuple):
            collectionType = c.TYPE_TUPLE
        elif isinstance(outterValue, dict) :
            collectionType = c.TYPE_DICT
        else :
            raise Exception(f'Unexpected collection: {outterValue}')
        return StringHelperHelper.prettyCollection(
            StringHelperHelper.getValueCollection(outterValue),
            collectionType,
            quote,
            prettyPython,
            tabCount,
            nullValue,
            trueValue,
            falseValue,
            withColors = withColors,
            joinAtReturn = joinAtReturn
        )
    else :
        return StringHelperHelper.prettyInstance(
            outterValue,
            quote,
            prettyPython,
            tabCount,
            nullValue,
            trueValue,
            falseValue,
            withColors = withColors,
            joinAtReturn = joinAtReturn
        )

def prettyJson(
        outterValue,
        quote = c.DOUBLE_QUOTE,
        tabCount = 0,
        nullValue = c.NULL_VALUE,
        trueValue = c.TRUE_VALUE,
        falseValue = c.FALSE_VALUE,
        withColors = False,
        joinAtReturn = True
    ) :
    '''It always sort sets'''
    if ObjectHelper.isCollection(outterValue) :
        if isinstance(outterValue, list) or isinstance(outterValue, set) or isinstance(outterValue, tuple) :
            collectionType = c.TYPE_LIST
        elif isinstance(outterValue, dict) :
            collectionType = c.TYPE_DICT
        else :
            raise Exception(f'Unexpected collection: {outterValue}')
        return StringHelperHelper.prettyCollection(
            StringHelperHelper.getValueCollection(outterValue),
            collectionType,
            quote,
            prettyJson,
            tabCount,
            nullValue,
            trueValue,
            falseValue,
            withColors = withColors,
            joinAtReturn = joinAtReturn
        )
    else :
        return StringHelperHelper.prettyInstance(
            outterValue,
            quote,
            prettyJson,
            tabCount,
            nullValue,
            trueValue,
            falseValue,
            withColors = withColors,
            joinAtReturn = joinAtReturn
        )

def filterString(string) :
    if ObjectHelper.isNone(string) or not isinstance(string, str) :
        return string
    strippedString = string.strip()
    if c.BLANK == strippedString :
        return strippedString
    if strippedString[-1] == c.NEW_LINE :
        strippedString = strippedString[:-1]
    surroundedBySingleQuote = strippedString.startswith(c.SINGLE_QUOTE) and strippedString.endswith(c.SINGLE_QUOTE) and not (c.SINGLE_QUOTE*2 == strippedString or c.TRIPLE_SINGLE_QUOTE == strippedString)
    surroundedByDoubleQuote = strippedString.startswith(c.DOUBLE_QUOTE) and strippedString.endswith(c.DOUBLE_QUOTE) and not (c.DOUBLE_QUOTE*2 == strippedString or c.TRIPLE_DOUBLE_QUOTE == strippedString)
    if c.HASH_TAG in strippedString and not (surroundedBySingleQuote or surroundedByDoubleQuote) :
        strippedString = filterString(c.HASH_TAG.join(strippedString.split(c.HASH_TAG)[:-1]))
    if strippedString and (
            (
                strippedString.startswith(c.TRIPLE_SINGLE_QUOTE) and
                strippedString.endswith(c.TRIPLE_SINGLE_QUOTE) and
                not c.TRIPLE_SINGLE_QUOTE == strippedString
            ) or
            (
                strippedString.startswith(c.TRIPLE_DOUBLE_QUOTE) and
                strippedString.endswith(c.TRIPLE_DOUBLE_QUOTE) and
                not c.TRIPLE_DOUBLE_QUOTE == strippedString
            )
        ) :
        return filterString(strippedString[3:-3])
    else :
        return strippedString if not (surroundedBySingleQuote or surroundedByDoubleQuote) else filterString(strippedString[1:-1])

def isLongString(thing) :
    if not thing is None and isinstance(thing, str) :
        return (
            (
                thing.startswith(c.TRIPLE_SINGLE_QUOTE) or
                thing.endswith(c.TRIPLE_SINGLE_QUOTE)
            ) and StringHelperHelper.isNotOneLineLongString(c.TRIPLE_SINGLE_QUOTE, thing) or (
                thing.startswith(c.TRIPLE_DOUBLE_QUOTE) or
                thing.endswith(c.TRIPLE_DOUBLE_QUOTE)
            ) and StringHelperHelper.isNotOneLineLongString(c.TRIPLE_DOUBLE_QUOTE, thing)
        )
    else :
        return False

def removeColors(thing) :
    if ObjectHelper.isNotNone(thing) and isNotBlank(thing) :
        for color in c.IMPLEMENTED_PROMP_COLORS :
            if color in thing :
                thing = thing.replace(color,c.BLANK)
    return thing if isinstance(thing, str) else str(string)

def getS(condition, es=False) :
    return c.BLANK if not condition else 'es' if es else 's'

def getToBe(condition, singular=True, tense=PRESENT, negative=False) :
    if condition :
        if negative :
            if PAST == tense :
                return f'wasn{c.SINGLE_QUOTE}t' if singular else f'weren{c.SINGLE_QUOTE}t'
            elif PRESENT == tense :
                return f'isn{c.SINGLE_QUOTE}t' if singular else f'aren{c.SINGLE_QUOTE}t'
            elif FUTURE == tense :
                return f'won{c.SINGLE_QUOTE}t'
        else :
            if PAST == tense :
                return 'was' if singular else 'were'
            elif PRESENT == tense :
                return 'is' if singular else 'are'
            elif FUTURE == tense :
                return 'will'
        return 'to be'
    else :
        return c.BLANK

def getOnlyLetters(value):
    return join([
        char for char in join(
            str(value).replace(c.UNDERSCORE, c.SPACE).replace(c.DASH, c.SPACE).split(),
            character=c.SPACE
        ) if char in c.CHARACTERES
    ])

def isUp(value):
    return 1 == len(str(value)) and str(value) in c.UPPER_CASE_CHARACTERES

def fromPascalToTitle(value):
    return join([f'{c.SPACE}{char}' if isUp(char) else char for char in str(value)], character=c.BLANK)

def toTitle(value):
    return join(fromPascalToTitle(getOnlyLetters(value)).title().split(), character=c.SPACE)

def toPascalCase(value):
    return join(toTitle(value).split())

def toCamelCase(value):
    content = toPascalCase(value)
    return join([content[:1].lower(), content[1:]])

def toSnakeCase(value):
    return join(toTitle(value).lower().split(), character=c.UNDERSCORE)

def toKebabCase(value):
    return join(toTitle(value).lower().split(), character=c.DASH)



from python_helper import SettingHelper, StringHelper, Constant, log, EnvironmentHelper, ObjectHelper, Test, RandomHelper

UNMUTED_LOG_HELPER_SETTINGS = {
    log.LOG : True,
    log.SUCCESS : True,
    log.SETTING : True,
    log.DEBUG : True,
    log.WARNING : True,
    log.WRAPPER : True,
    log.FAILURE : True,
    log.ERROR : True,
    log.TEST : False
}

MUTED_LOG_HELPER_SETTINGS = {
    log.LOG : False,
    log.SUCCESS : False,
    log.SETTING : False,
    log.DEBUG : False,
    log.WARNING : False,
    log.WRAPPER : False,
    log.FAILURE : False,
    log.ERROR : False,
    log.TEST : False
}

TEST_SETTINGS = {
    'inspectGlobals' : False,
    'logResult' : True
}

@Test(
    environmentVariables = {
        log.ENABLE_LOGS_WITH_COLORS: True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **UNMUTED_LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def mustRun_withSuccess() :
    # Arrange
    def myFunction(a):
        return a
    @Test(environmentVariables={
        SettingHelper.ACTIVE_ENVIRONMENT : None,
        **MUTED_LOG_HELPER_SETTINGS
    })
    def myTest() :
        a = 'original a'
        originalA = str(a) + ''
        b = myFunction(a)
        assert originalA == b
    exception = None

    # Act
    try :
        myTest()
    except Exception as e :
        exception = e

    # Assert
    assert myFunction('original a') == str('original a')+''
    assert exception is None

@Test(
    environmentVariables = {
        log.ENABLE_LOGS_WITH_COLORS: True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **UNMUTED_LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def mustRun_withFailure() :
    # Arrange
    def myFunction(a):
        return a
    @Test(environmentVariables={
        SettingHelper.ACTIVE_ENVIRONMENT : None,
        **MUTED_LOG_HELPER_SETTINGS
    })
    def myTest() :
        assert 'b' == 'a'
    exception = None

    # Act
    try :
        myTest()
    except Exception as e :
        exception = e

    # Assert
    assert exception is not None

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS: True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **UNMUTED_LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def mustRun_withSuccess_ExecutingActionFirst() :
    # Arrange
    def myAction(c, d=None):
        return c+d
    def myFunction(a):
        return a
    returns = {}
    @Test(
        callBefore=myAction,
        argsOfCallBefore='f',
        kwargsOfCallBefore={'d':'g'},
        returns=returns,
        environmentVariables={
            SettingHelper.ACTIVE_ENVIRONMENT : None,
            **MUTED_LOG_HELPER_SETTINGS
        }
    )
    def myTest() :
        assert 'a' == myFunction('a')
    exception = None

    # Act
    try :
        myTest()
    except Exception as e :
        exception = e

    # Assert
    assert ObjectHelper.isNone(exception)
    assert 'fg' == returns['returnOfCallBefore']

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS: True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **UNMUTED_LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def mustRun_withFailre_ExecutingActionFirst() :
    # Arrange
    def myAction(c, d=None):
        return c+d
    def myFunction(a):
        return a
    returns = {}
    @Test(
        callBefore=myAction,
        argsOfCallBefore='f',
        kwargsOfCallBefore={'d':'g'},
        returns=returns,
        environmentVariables={
            SettingHelper.ACTIVE_ENVIRONMENT : None,
            **MUTED_LOG_HELPER_SETTINGS
        }
    )
    def myTest() :
        assert 'a' == myFunction('b')
    exception = None

    # Act
    try :
        myTest()
    except Exception as e :
        exception = e

    # Assert
    assert ObjectHelper.isNotNone(exception)
    assert 'fg' == returns['returnOfCallBefore']

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS: True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **UNMUTED_LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def mustRun_withSuccess_ExecutingActionFirst_withFailure() :
    # Arrange
    exceptionMessage = 'some exception'
    def myAction(c, d=None):
        raise Exception(exceptionMessage)
    def myFunction(a):
        return a
    returns = {}
    @Test(
        callBefore=myAction,
        argsOfCallBefore='f',
        kwargsOfCallBefore={'d':'g'},
        returns=returns,
        environmentVariables={
            SettingHelper.ACTIVE_ENVIRONMENT : None,
            **MUTED_LOG_HELPER_SETTINGS
        }
    )
    def myTest() :
        assert 'a' == myFunction('a')
    exception = None

    # Act
    try :
        myTest()
    except Exception as e :
        exception = e

    # Assert
    assert ObjectHelper.isNotNone(exception)
    assert exceptionMessage == str(exception)

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS: True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **UNMUTED_LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def mustRun_withFailre_ExecutingActionFirst_withFailre() :
    # Arrange
    exceptionMessage = 'some exception'
    def myAction(c, d=None):
        raise Exception(exceptionMessage)
    def myFunction(a):
        return a
    returns = {}
    @Test(
        callBefore=myAction,
        argsOfCallBefore='f',
        kwargsOfCallBefore={'d':'g'},
        returns=returns,
        environmentVariables={
            SettingHelper.ACTIVE_ENVIRONMENT : None,
            **MUTED_LOG_HELPER_SETTINGS
        }
    )
    def myTest() :
        assert 'a' == myFunction('b')
    exception = None

    # Act
    try :
        myTest()
    except Exception as e :
        exception = e

    # Assert
    assert ObjectHelper.isNotNone(exception)
    assert exceptionMessage == str(exception)

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS: True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **UNMUTED_LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def mustRun_withSuccess_ExecutingActionLater() :
    # Arrange
    def myAction(c, d=None):
        return c+d
    def myFunction(a):
        return a
    returns = {}
    @Test(
        callAfter=myAction,
        argsOfCallAfter='f',
        kwargsOfCallAfter={'d':'g'},
        returns=returns,
        environmentVariables={
            SettingHelper.ACTIVE_ENVIRONMENT : None,
            **MUTED_LOG_HELPER_SETTINGS
        }
    )
    def myTest() :
        assert 'a' == myFunction('a')
    exception = None

    # Act
    try :
        myTest()
    except Exception as e :
        exception = e

    # Assert
    assert ObjectHelper.isNone(exception)
    assert 'fg' == returns['returnOfCallAfter']

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS: True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **UNMUTED_LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def mustRun_withFailre_ExecutingActionLater() :
    # Arrange
    def myAction(c, d=None):
        return c+d
    def myFunction(a):
        return a
    returns = {}
    @Test(
        callAfter=myAction,
        argsOfCallAfter='f',
        kwargsOfCallAfter={'d':'g'},
        returns=returns,
        environmentVariables={
            SettingHelper.ACTIVE_ENVIRONMENT : None,
            **MUTED_LOG_HELPER_SETTINGS
        }
    )
    def myTest() :
        assert 'a' == myFunction('b')
    exception = None

    # Act
    try :
        myTest()
    except Exception as e :
        exception = e

    # Assert
    assert ObjectHelper.isNotNone(exception)
    assert 'fg' == returns['returnOfCallAfter']

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS: True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **UNMUTED_LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def mustRun_withSuccess_ExecutingActionLater_withFailure() :
    # Arrange
    exceptionMessage = 'some exception'
    def myAction(c, d=None):
        raise Exception(exceptionMessage)
    def myFunction(a):
        return a
    returns = {}
    @Test(
        callAfter=myAction,
        argsOfCallAfter='f',
        kwargsOfCallAfter={'d':'g'},
        returns=returns,
        environmentVariables={
            SettingHelper.ACTIVE_ENVIRONMENT : None,
            **MUTED_LOG_HELPER_SETTINGS
        }
    )
    def myTest() :
        assert 'a' == myFunction('a')
    exception = None

    # Act
    try :
        myTest()
    except Exception as e :
        exception = e

    # Assert
    assert ObjectHelper.isNotNone(exception)
    assert exceptionMessage == str(exception)

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS: True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **UNMUTED_LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def mustRun_withFailre_ExecutingActionLater_withFailre() :
    # Arrange
    exceptionMessage = 'some exception'
    def myAction(c, d=None):
        raise Exception(exceptionMessage)
    def myFunction(a):
        return a
    returns = {}
    @Test(
        callAfter=myAction,
        argsOfCallAfter='f',
        kwargsOfCallAfter={'d':'g'},
        returns=returns,
        environmentVariables={
            SettingHelper.ACTIVE_ENVIRONMENT : None,
            **MUTED_LOG_HELPER_SETTINGS
        }
    )
    def myTest() :
        assert 'a' == myFunction('b')
    exception = None

    # Act
    try :
        myTest()
    except Exception as e :
        exception = e

    # Assert
    assert ObjectHelper.isNotNone(exception)
    assert not exceptionMessage == str(exception)
    assert 'AssertionError. Followed by: some exception' == str(exception)

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS: True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **UNMUTED_LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def handleEnvironmentChangesProperly_withSuccess() :
    beforeTestEnvironmentSettings = {**EnvironmentHelper.getSet()}
    inBetweenTestEnvironmentSettings = None
    afterTestEnvironmentSettings = None
    # Arrange
    def myBeforeAction(c, d=None):
        return d+c
    def myAfterAction(c, d=None):
        return c+d
    def myFunction(a):
        return a
    returns = {}
    @Test(
        callBefore=myBeforeAction,
        argsOfCallBefore='f',
        kwargsOfCallBefore={'d':'g'},
        callAfter=myAfterAction,
        argsOfCallAfter='h',
        kwargsOfCallAfter={'d':'i'},
        returns=returns,
        environmentVariables={
            SettingHelper.ACTIVE_ENVIRONMENT : None,
            **MUTED_LOG_HELPER_SETTINGS
        }
    )
    def myTest() :
        inBetweenTestEnvironmentSettings = EnvironmentHelper.getSet()
        assert ObjectHelper.isNotNone(inBetweenTestEnvironmentSettings)
        assert ObjectHelper.isDictionary(inBetweenTestEnvironmentSettings)
        assert 'a' == myFunction('a')
        return inBetweenTestEnvironmentSettings

    # Act
    inBetweenTestEnvironmentSettings = myTest()
    afterTestEnvironmentSettings = {**EnvironmentHelper.getSet()}

    # Assert
    assert 'gf' == returns['returnOfCallBefore']
    assert 'hi' == returns['returnOfCallAfter']
    assert ObjectHelper.isNotNone(beforeTestEnvironmentSettings)
    assert ObjectHelper.isDictionary(beforeTestEnvironmentSettings)
    assert ObjectHelper.isNotNone(inBetweenTestEnvironmentSettings)
    assert ObjectHelper.isDictionary(inBetweenTestEnvironmentSettings)
    assert ObjectHelper.isNotNone(afterTestEnvironmentSettings)
    assert ObjectHelper.isDictionary(afterTestEnvironmentSettings)
    assert afterTestEnvironmentSettings == beforeTestEnvironmentSettings
    assert not beforeTestEnvironmentSettings == inBetweenTestEnvironmentSettings


@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS: True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **UNMUTED_LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def handleEnvironmentChangesProperly_withSuccess_whenActionsHaveNoArguments() :
    beforeTestEnvironmentSettings = {**EnvironmentHelper.getSet()}
    inBetweenTestEnvironmentSettings = None
    afterTestEnvironmentSettings = None
    MY_BEFORE_ACTION_RETURN = RandomHelper.string(minimum=10)
    MY_AFTER_ACTION_RETURN = RandomHelper.string(minimum=10)
    # Arrange
    def myBeforeAction():
        return MY_BEFORE_ACTION_RETURN
    def myAfterAction():
        return MY_AFTER_ACTION_RETURN
    def myFunction(a):
        return a
    returns = {}
    @Test(
        callBefore=myBeforeAction,
        callAfter=myAfterAction,
        returns=returns,
        environmentVariables={
            SettingHelper.ACTIVE_ENVIRONMENT : None,
            **MUTED_LOG_HELPER_SETTINGS
        }
    )
    def myTest() :
        inBetweenTestEnvironmentSettings = EnvironmentHelper.getSet()
        assert ObjectHelper.isNotNone(inBetweenTestEnvironmentSettings)
        assert ObjectHelper.isDictionary(inBetweenTestEnvironmentSettings)
        assert 'a' == myFunction('a')
        return inBetweenTestEnvironmentSettings

    # Act
    inBetweenTestEnvironmentSettings = myTest()
    afterTestEnvironmentSettings = {**EnvironmentHelper.getSet()}

    # Assert
    assert MY_BEFORE_ACTION_RETURN == returns['returnOfCallBefore']
    assert MY_AFTER_ACTION_RETURN == returns['returnOfCallAfter']
    assert ObjectHelper.isNotNone(beforeTestEnvironmentSettings)
    assert ObjectHelper.isDictionary(beforeTestEnvironmentSettings)
    assert ObjectHelper.isNotNone(inBetweenTestEnvironmentSettings)
    assert ObjectHelper.isDictionary(inBetweenTestEnvironmentSettings)
    assert ObjectHelper.isNotNone(afterTestEnvironmentSettings)
    assert ObjectHelper.isDictionary(afterTestEnvironmentSettings)
    assert afterTestEnvironmentSettings == beforeTestEnvironmentSettings
    assert not beforeTestEnvironmentSettings == inBetweenTestEnvironmentSettings


@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS: True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **UNMUTED_LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def handleEnvironmentChangesProperly_withErrorBefore() :
    beforeTestEnvironmentSettings = {**EnvironmentHelper.getSet()}
    inBetweenTestEnvironmentSettings = None
    afterTestEnvironmentSettings = None
    # Arrange
    exceptionMessage = 'some exception'
    def myBeforeAction(c, d=None):
        raise Exception(exceptionMessage)
    def myAfterAction(c, d=None):
        return c+d
    def myFunction(a):
        return a
    returns = {}
    exception = None
    @Test(
        callBefore=myBeforeAction,
        argsOfCallBefore='f',
        kwargsOfCallBefore={'d':'g'},
        callAfter=myAfterAction,
        argsOfCallAfter='h',
        kwargsOfCallAfter={'d':'i'},
        returns=returns,
        environmentVariables={
            SettingHelper.ACTIVE_ENVIRONMENT : None,
            **MUTED_LOG_HELPER_SETTINGS
        }
    )
    def myTest() :
        inBetweenTestEnvironmentSettings = EnvironmentHelper.getSet()
        assert ObjectHelper.isNotNone(inBetweenTestEnvironmentSettings)
        assert ObjectHelper.isDictionary(inBetweenTestEnvironmentSettings)
        assert 'a' == myFunction('a')
        return inBetweenTestEnvironmentSettings

    # Act
    try :
        inBetweenTestEnvironmentSettings = myTest()
    except Exception as e :
        exception = e
    afterTestEnvironmentSettings = {**EnvironmentHelper.getSet()}

    assert {} == returns
    assert ObjectHelper.isNotEmpty(exception)
    assert exceptionMessage == str(exception)
    assert ObjectHelper.isNotNone(beforeTestEnvironmentSettings)
    assert ObjectHelper.isDictionary(beforeTestEnvironmentSettings)
    assert ObjectHelper.isNone(inBetweenTestEnvironmentSettings)
    assert ObjectHelper.isNotNone(afterTestEnvironmentSettings)
    assert ObjectHelper.isDictionary(afterTestEnvironmentSettings)
    assert afterTestEnvironmentSettings == beforeTestEnvironmentSettings
    assert not beforeTestEnvironmentSettings == inBetweenTestEnvironmentSettings

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS: True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **UNMUTED_LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def handleEnvironmentChangesProperly_withErrorAfter() :
    beforeTestEnvironmentSettings = {**EnvironmentHelper.getSet()}
    inBetweenTestEnvironmentSettings = None
    afterTestEnvironmentSettings = None
    # Arrange
    exceptionMessage = 'some exception'
    def myBeforeAction(c, d=None):
        return d+c
    def myAfterAction(c, d=None):
        raise Exception(exceptionMessage)
    def myFunction(a):
        return a
    returns = {}
    exception = None
    @Test(
        callBefore=myBeforeAction,
        argsOfCallBefore='f',
        kwargsOfCallBefore={'d':'g'},
        callAfter=myAfterAction,
        argsOfCallAfter='h',
        kwargsOfCallAfter={'d':'i'},
        returns=returns,
        environmentVariables={
            SettingHelper.ACTIVE_ENVIRONMENT : None,
            **MUTED_LOG_HELPER_SETTINGS
        }
    )
    def myTest() :
        inBetweenTestEnvironmentSettings = EnvironmentHelper.getSet()
        assert ObjectHelper.isNotNone(inBetweenTestEnvironmentSettings)
        assert ObjectHelper.isDictionary(inBetweenTestEnvironmentSettings)
        assert 'a' == myFunction('a')
        return inBetweenTestEnvironmentSettings

    # Act
    try :
        inBetweenTestEnvironmentSettings = myTest()
    except Exception as e :
        exception = e
    afterTestEnvironmentSettings = {**EnvironmentHelper.getSet()}

    assert ObjectHelper.isNotEmpty(exception)
    assert exceptionMessage == str(exception)
    assert 'gf' == returns['returnOfCallBefore']
    assert {'returnOfCallBefore':'gf'} == returns
    assert ObjectHelper.isNotNone(beforeTestEnvironmentSettings)
    assert ObjectHelper.isDictionary(beforeTestEnvironmentSettings)
    assert ObjectHelper.isNone(inBetweenTestEnvironmentSettings)
    assert ObjectHelper.isNotNone(afterTestEnvironmentSettings)
    assert ObjectHelper.isDictionary(afterTestEnvironmentSettings)
    assert afterTestEnvironmentSettings == beforeTestEnvironmentSettings
    assert not beforeTestEnvironmentSettings == inBetweenTestEnvironmentSettings

@Test(
    environmentVariables={
        log.ENABLE_LOGS_WITH_COLORS: True,
        SettingHelper.ACTIVE_ENVIRONMENT : SettingHelper.LOCAL_ENVIRONMENT,
        **UNMUTED_LOG_HELPER_SETTINGS
    },
    **TEST_SETTINGS
)
def handleEnvironmentChangesProperly_withError() :
    beforeTestEnvironmentSettings = {**EnvironmentHelper.getSet()}
    inBetweenTestEnvironmentSettings = None
    afterTestEnvironmentSettings = None
    someExceptionMessage = 'some exception message'
    # Arrange
    def myBeforeAction(c, d=None):
        return d+c
    def myAfterAction(c, d=None):
        return c+d
    def myFunction(a):
        raise Exception(someExceptionMessage)
    returns = {}
    @Test(
        callBefore=myBeforeAction,
        argsOfCallBefore='f',
        kwargsOfCallBefore={'d':'g'},
        callAfter=myAfterAction,
        argsOfCallAfter='h',
        kwargsOfCallAfter={'d':'i'},
        returns=returns,
        environmentVariables={
            SettingHelper.ACTIVE_ENVIRONMENT : None,
            **MUTED_LOG_HELPER_SETTINGS
        }
    )
    def myTest() :
        inBetweenTestEnvironmentSettings = EnvironmentHelper.getSet()
        assert ObjectHelper.isNotNone(inBetweenTestEnvironmentSettings)
        assert ObjectHelper.isDictionary(inBetweenTestEnvironmentSettings)
        assert 'a' == myFunction('a')
        return inBetweenTestEnvironmentSettings

    # Act
    try :
        inBetweenTestEnvironmentSettings = myTest()
    except Exception as e :
        exception = e
    afterTestEnvironmentSettings = {**EnvironmentHelper.getSet()}

    # Assert
    assert 'gf' == returns['returnOfCallBefore']
    assert 'hi' == returns['returnOfCallAfter']
    assert ObjectHelper.isNotEmpty(exception)
    assert someExceptionMessage == str(exception)
    assert ObjectHelper.isNotNone(beforeTestEnvironmentSettings)
    assert ObjectHelper.isDictionary(beforeTestEnvironmentSettings)
    assert ObjectHelper.isNone(inBetweenTestEnvironmentSettings)
    assert ObjectHelper.isNotNone(afterTestEnvironmentSettings)
    assert ObjectHelper.isDictionary(afterTestEnvironmentSettings)
    assert afterTestEnvironmentSettings == beforeTestEnvironmentSettings
    assert not beforeTestEnvironmentSettings == inBetweenTestEnvironmentSettings
